"""Chop and unchop

https://github.com/tidyverse/tidyr/blob/master/R/chop.R
"""
from collections import defaultdict
from typing import Iterable, List, Mapping, Tuple, Union

import numpy
import pandas
from pandas import DataFrame, Series
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.types import IntOrIter, StringOrIter, Dtype, is_scalar
from ..core.utils import (
    df_getitem,
    df_setitem,
    vars_select,
    copy_attrs,
    apply_dtypes,
    keep_column_order,
    reconstruct_tibble,
)
from ..core.exceptions import ColumnNotExistingError
from ..core.contexts import Context
from ..core.grouped import DataFrameGroupBy

from ..base import union, NA
from ..dplyr import bind_cols, group_by, arrange, group_data

from .drop_na import drop_na


@register_verb(DataFrame, context=Context.SELECT)
def chop(
    data: DataFrame,
    cols: Union[IntOrIter, StringOrIter] = None,
    base0_: bool = None,
) -> DataFrame:
    """Makes data frame shorter by converting rows within each group
    into list-columns.

    Args:
        data: A data frame
        cols: Columns to chop
        base0_: Whether `cols` are 0-based
            if not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        Data frame with selected columns chopped
    """
    if cols is None:
        return data.copy()

    all_columns = data.columns
    cols = vars_select(all_columns, cols, base0=base0_)
    cols = all_columns[cols]
    # when cols is empty
    # order may change for all_columns.difference([])
    key_cols = all_columns.difference(cols) if len(cols) > 0 else all_columns

    vals = data[cols]
    keys = data[key_cols]

    compacted = []
    if data.shape[0] == 0:
        split_key = keys
    else:
        split = _vec_split(vals, keys)
        try:
            split_key = df_getitem(split, "key")
        except (KeyError, ColumnNotExistingError):
            split_key = None
        split_val = df_getitem(split, "val")

        for val in split_val:
            compacted.append(_compact_df(val))

    if not compacted:
        vals = DataFrame(columns=cols)
    else:
        vals = pandas.concat(compacted, ignore_index=True)

    out = bind_cols(split_key, vals, __calling_env=CallingEnvs.REGULAR)
    return reconstruct_tibble(data, out, keep_rowwise=True)


@register_verb(DataFrame, context=Context.SELECT)
def unchop(
    data: DataFrame,
    cols: Union[IntOrIter, StringOrIter] = None,
    keep_empty: bool = False,
    ptype: Union[Dtype, Mapping[str, Dtype]] = None,
    base0_: bool = None,
) -> DataFrame:
    """Makes df longer by expanding list-columns so that each element
    of the list-column gets its own row in the output.

    See https://tidyr.tidyverse.org/reference/chop.html

    Recycling size-1 elements might be different from `tidyr`
        >>> df = tibble(x=[1, [2,3]], y=[[2,3], 1])
        >>> df >> unchop([f.x, f.y])
        >>> # tibble(x=[1,2,3], y=[2,3,1])
        >>> # instead of following in tidyr
        >>> # tibble(x=[1,1,2,3], y=[2,3,1,1])

    Args:
        data: A data frame.
        cols: Columns to unchop.
        keep_empty: By default, you get one row of output for each element
            of the list your unchopping/unnesting.
            This means that if there's a size-0 element
            (like NULL or an empty data frame), that entire row will be
            dropped from the output.
            If you want to preserve all rows, use `keep_empty` = `True` to
            replace size-0 elements with a single row of missing values.
        ptype: Providing the dtypes for the output columns.
            Could be a single dtype, which will be applied to all columns, or
            a dictionary of dtypes with keys for the columns and values the
            dtypes.
            For nested data frames, we need to specify `col$a` as key. If `col`
            is used as key, all columns of the nested data frames will be casted
            into that dtype.
        base0_: Whether `cols` are 0-based
            if not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        A data frame with selected columns unchopped.
    """
    all_columns = data.columns
    cols = vars_select(all_columns, cols, base0=base0_)

    if len(cols) == 0 or data.shape[0] == 0:
        return (
            data.copy(copy_grouped=True)
            if isinstance(data, DataFrameGroupBy)
            else data.copy()
        )

    cols = all_columns[cols]
    key_cols = all_columns.difference(cols).tolist()
    out = _unchopping(data, cols, key_cols, keep_empty)

    apply_dtypes(out, ptype)
    return reconstruct_tibble(data, out, keep_rowwise=True)


def _vec_split(
    x: Union[DataFrame, Series], by: Union[DataFrame, Series]
) -> DataFrame:
    """Split a vector into groups

    Returns a data frame with columns `key` and `val`. `key` is a stacked column
    with data from by.
    """
    if isinstance(x, Series):  # pragma: no cover, always a data frame?
        x = x.to_frame()
    if isinstance(by, Series):  # pragma: no cover, always a data frame?
        by = by.to_frame()

    df = bind_cols(x, by, __calling_env=CallingEnvs.REGULAR)
    if df.shape[0] == 0:
        return DataFrame(columns=["key", "val"])
    df = group_by(df, *by.columns, __calling_env=CallingEnvs.REGULAR)
    gdata = group_data(df, __calling_env=CallingEnvs.REGULAR)
    gdata = arrange(gdata, gdata._rows, __calling_env=CallingEnvs.REGULAR)
    out = DataFrame(index=gdata.index)
    out = df_setitem(out, "key", gdata[by.columns])
    return df_setitem(
        out,
        "val",
        [x.iloc[rows, :].reset_index(drop=True) for rows in gdata._rows],
    )


def _compact_df(data: DataFrame) -> DataFrame:
    """Compact each series as list in a data frame"""
    out = DataFrame(index=[0], columns=data.columns)
    for col in data.columns:
        out.loc[0, col] = data[col].values.tolist()
    return out


def _unchopping(
    data: DataFrame,
    data_cols: Iterable[str],
    key_cols: Iterable[str],
    keep_empty: bool,
) -> DataFrame:
    """Unchop the data frame

    See https://stackoverflow.com/questions/53218931/how-to-unnest-explode-a-column-in-a-pandas-dataframe
    """  # noqa: E501
    # key_cols could be empty
    rsize = None
    val_data = {}
    for dcol in data_cols:
        # check dtype first so that we don't need to check
        # other types of columns element by element
        is_df_col = data_cols.dtype == object and all(
            # it's either null or a dataframe
            (is_scalar(val) and pandas.isnull(val))
            or isinstance(val, DataFrame)
            for val in data[dcol]
        )
        if is_df_col:
            vdata, sizes, dtypes = _unchopping_df_column(data[dcol])
        else:
            vdata, sizes, dtypes = _unchopping_nondf_column(data[dcol])
        val_data.update(vdata)

        if rsize is None:
            rsize = sizes
        else:
            tmpsize = []
            for prevsize, cursize in zip(rsize, sizes):
                if prevsize != cursize and 1 not in (prevsize, cursize):
                    raise ValueError(
                        f"Incompatible lengths: {prevsize}, {cursize}."
                    )
                tmpsize.append(max(prevsize, cursize))
            rsize = tmpsize

    key_data = {key: numpy.repeat(data[key].values, rsize) for key in key_cols}
    key_data.update(val_data)
    # DataFrame(key_data) may have nested dfs
    # say y$a, then ['y'] will not select it
    out = keep_column_order(DataFrame(key_data), data.columns)
    if not keep_empty:
        out = drop_na(
            out,
            *val_data,
            how_="all",
            __calling_env=CallingEnvs.REGULAR,
        )
    apply_dtypes(out, dtypes)
    copy_attrs(out, data)
    return out


def _unchopping_df_column(
    series: Series,
) -> Tuple[Mapping[str, List], List[int], Mapping[str, Dtype]]:
    """Unchopping dataframe column"""
    # Get union column names
    union_cols = []
    # try to keep the same dtype
    dtypes = None
    for val in series:
        if isinstance(val, DataFrame):
            union_cols = union(
                union_cols,
                val.columns,
                __calling_env=CallingEnvs.REGULAR,
            )
            if dtypes is None:
                dtypes = {col: val[col].dtype for col in val}
            else:
                for col in val:
                    # TODO: test
                    if col in dtypes and dtypes[col] != val[col].dtype:
                        del dtypes[col]  # pragma: no cover

    sizes = []
    val_data = defaultdict(list)
    # add missing columns to each df
    for val in series:
        if isinstance(val, DataFrame):
            for col in union_cols:
                val_data[f"{series.name}${col}"].extend(
                    val[col] if col in val else [NA] * val.shape[0]
                )
            sizes.append(val.shape[0])
        else:  # null
            for col in union_cols:
                val_data[f"{series.name}${col}"].append(NA)
            sizes.append(1)

    return val_data, sizes, dtypes


def _unchopping_nondf_column(
    series: Series,
) -> Tuple[Mapping[str, List], List[int], Mapping[str, Dtype]]:
    """Unchopping non-dataframe column"""
    val_data = {}
    vals = [[val] if is_scalar(val) else val for val in series]
    val_data[series.name] = Series(
        numpy.concatenate(
            vals,
            axis=None,
            # casting="no" # only for numpy 1.20.0+
        ),
        dtype=series.dtype,
    )
    return val_data, [len(val) for val in vals], {}



"""Provides forcats verbs to manipulate factor level orders"""
from typing import Any, Callable, Iterable, Sequence

from pandas import Categorical, DataFrame, Series
from pipda import register_func, register_verb
from pipda.utils import CallingEnvs, functype

from ..base import (
    NA,
    append,
    as_integer,
    duplicated,
    intersect,
    levels,
    match,
    median,
    nlevels,
    order,
    rev,
    sample,
    seq_len,
    setdiff,
    table
)
from ..core.contexts import Context
from ..core.types import (
    ForcatsRegType,
    ForcatsType,
    is_not_null,
    is_null,
    is_scalar
)
from ..core.utils import get_option, logger
from .lvls import lvls_reorder, lvls_seq
from .utils import check_factor


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_relevel(
    _f: ForcatsType,
    *lvls: Any,
    after: int = 0,
    base0_: bool = None,
) -> Categorical:
    """Reorder factor levels by hand

    Args:
        _f: A factor (categoriccal), or a string vector
        *lvls: Either a function (then `len(lvls)` should equal to `1`) or
            the new levels.
            A function will be called with the current levels as input, and the
            return value (which must be a character vector) will be used to
            relevel the factor.
            Any levels not mentioned will be left in their existing order,
            by default after the explicitly mentioned levels.
        after: Where should the new values be placed?
        base0_: Whether after is 0-based.
            if not given, will use `get_option("index.base.0")`.
            When it's 1-based, after=0 will append to the beginning,
            -1 will append to the end.
            When 0-based, after=None will append to the beginning,
            -1 to the end

    Returns:
        The factor with levels replaced
    """

    _f = check_factor(_f)
    old_levels = levels(_f, __calling_env=CallingEnvs.REGULAR)
    if len(lvls) == 1 and callable(lvls[0]):
        first_levels = lvls[0](old_levels)
    else:
        first_levels = lvls

    unknown = setdiff(
        first_levels,
        old_levels,
        __calling_env=CallingEnvs.REGULAR,
    )

    if len(unknown) > 0:
        logger.warning("[fct_relevel] Unknown levels in `_f`: %s", unknown)
        first_levels = intersect(
            first_levels,
            old_levels,
            __calling_env=CallingEnvs.REGULAR,
        )

    new_levels = append(
        setdiff(old_levels, first_levels, __calling_env=CallingEnvs.REGULAR),
        first_levels,
        after=after,
        __calling_env=CallingEnvs.REGULAR,
    )

    base0_ = get_option("index.base.0", base0_)
    return lvls_reorder(
        _f,
        match(  # base0_ defaults to which.base.0
            new_levels,
            old_levels,
            base0_=base0_,
            __calling_env=CallingEnvs.REGULAR,
        ),
        base0_=base0_,
        __calling_env=CallingEnvs.REGULAR,
    )


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_inorder(_f: ForcatsType, ordered: bool = None) -> Categorical:
    """Reorder factor levels by first appearance

    Args:
        _f: A factor
        ordered: A logical which determines the "ordered" status of the
            output factor.

    Returns:
        The factor with levels reordered
    """
    _f = check_factor(_f)
    dups = duplicated(_f, __calling_env=CallingEnvs.REGULAR)
    idx = as_integer(_f, __calling_env=CallingEnvs.REGULAR)[~dups]
    idx = idx[is_not_null(_f[~dups])]
    return lvls_reorder(
        _f,
        idx,
        ordered=ordered,
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


as_factor = fct_inorder


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_infreq(_f: ForcatsType, ordered: bool = None) -> Categorical:
    """Reorder factor levels by frequency

    Args:
        _f: A factor
        ordered: A logical which determines the "ordered" status of the
            output factor.

    Returns:
        The factor with levels reordered
    """
    _f = check_factor(_f)
    return lvls_reorder(
        _f,
        order(
            table(_f, __calling_env=CallingEnvs.REGULAR).values.flatten(),
            decreasing=True,
            base0_=True,
            __calling_env=CallingEnvs.REGULAR,
        ),
        ordered=ordered,
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_inseq(_f: ForcatsType, ordered: bool = None) -> Categorical:
    """Reorder factor levels by numeric order

    Args:
        _f: A factor
        ordered: A logical which determines the "ordered" status of the
            output factor.

    Returns:
        The factor with levels reordered
    """
    _f = check_factor(_f)
    levs = levels(_f, __calling_env=CallingEnvs.REGULAR)
    num_levels = []
    for lev in levs:
        try:
            numlev = as_integer(lev, __calling_env=CallingEnvs.REGULAR)
        except (ValueError, TypeError):
            numlev = NA
        num_levels.append(numlev)

    if all(is_null(num_levels)):
        raise ValueError(
            "At least one existing level must be coercible to numeric."
        )

    return lvls_reorder(
        _f,
        order(
            num_levels,
            na_last=True,
            base0_=True,
            __calling_env=CallingEnvs.REGULAR,
        ),
        ordered=ordered,
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


@register_func(None, context=Context.EVAL)
def last2(_x: Iterable, _y: Sequence) -> Any:
    """Find the last element of `_y` ordered by `_x`

    Args:
        _x: The vector used to order `_y`
        _y: The vector to get the last element of

    Returns:
        Last element of `_y` ordered by `_x`
    """
    return list(_y[order(_x, na_last=False, base0_=True)])[-1]


@register_func(None, context=Context.EVAL)
def first2(_x: Sequence, _y: Sequence) -> Any:
    """Find the first element of `_y` ordered by `_x`

    Args:
        _x: The vector used to order `_y`
        _y: The vector to get the first element of

    Returns:
        First element of `_y` ordered by `_x`
    """
    return _y[order(_x, base0_=True)][0]


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_reorder(
    _f: ForcatsType,
    _x: Sequence,
    *args: Any,
    _fun: Callable = median,
    _desc: bool = False,
    **kwargs: Any,
) -> Categorical:
    """Reorder factor levels by sorting along another variable

    Args:
        _f: A factor
        _x: The levels of `f` are reordered so that the values
            of `_fun(_x)` are in ascending order.
        _fun: The summary function, have to be passed by keyword
        *args, **kwargs: Other arguments for `_fun`.
        _desc: Order in descending order?

    Returns:
        The factor with levels reordered
    """
    _f = check_factor(_f)
    if is_scalar(_x):
        _x = [_x]

    if len(_f) != len(_x):
        raise ValueError("Unmatched length between `_x` and `_f`.")

    if functype(_fun) != "plain":
        kwargs["__calling_env"] = CallingEnvs.REGULAR

    args = args[1:]
    # simulate tapply
    summary = (
        DataFrame({"f": _f, "x": _x})
        .groupby("f", observed=False)
        .agg(lambda col: _fun(col, *args, **kwargs))
    )

    if not is_scalar(summary.iloc[0, 0]):
        raise ValueError("`fun` must return a single value per group.")

    return lvls_reorder(
        _f,
        order(
            summary.iloc[:, 0],
            decreasing=_desc,
            base0_=True,
            __calling_env=CallingEnvs.REGULAR,
        ),
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_reorder2(
    _f: ForcatsType,
    _x: Sequence,
    _y: Sequence,
    *args: Any,
    _fun: Callable = last2,
    _desc: bool = True,
    **kwargs: Any,
) -> Categorical:
    """Reorder factor levels by sorting along another variable

    Args:
        _f: A factor
        _x: and
        _y: The levels of `f` are reordered so that the values
            of `_fun(_x, _y)` are in ascending order.
        _fun: The summary function, have to be passed by keyword
        *args, **kwargs: Other arguments for `_fun`.
        _desc: Order in descending order?

    Returns:
        The factor with levels reordered
    """
    _f = check_factor(_f)
    if is_scalar(_x):
        _x = [_x]
    if is_scalar(_y):
        _y = [_y]
    if len(_f) != len(_x) or len(_f) != len(_y):
        raise ValueError("Unmatched length between `_x` and `_f`.")

    if functype(_fun) != "plain":
        kwargs["__calling_env"] = CallingEnvs.REGULAR

    args = args[1:]
    # simulate tapply
    summary = (
        DataFrame({"f": _f, "x": _x, "y": _y})
        .groupby("f", observed=False)
        .apply(
            lambda row: _fun(
                row.x.reset_index(drop=True),
                row.y.reset_index(drop=True),
                *args,
                **kwargs,
            )
        )
    )

    if not isinstance(summary, Series) or not is_scalar(summary[0]):
        raise ValueError("`fun` must return a single value per group.")

    return lvls_reorder(
        _f,
        order(
            summary,
            decreasing=_desc,
            base0_=True,
            __calling_env=CallingEnvs.REGULAR,
        ),
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


@register_verb(ForcatsRegType)
def fct_shuffle(_f: ForcatsType) -> Categorical:
    """Randomly permute factor levels

    Args:
        _f: A factor

    Returns:
        The factor with levels randomly permutated
    """
    _f = check_factor(_f)

    return lvls_reorder(
        _f,
        sample(lvls_seq(_f, base0_=True), __calling_env=CallingEnvs.REGULAR),
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


@register_verb(ForcatsRegType)
def fct_rev(_f: ForcatsType) -> Categorical:
    """Reverse order of factor levels

    Args:
        _f: A factor

    Returns:
        The factor with levels reversely ordered
    """
    _f = check_factor(_f)

    return lvls_reorder(
        _f,
        rev(lvls_seq(_f, base0_=True), __calling_env=CallingEnvs.REGULAR),
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_shift(_f: ForcatsType, n: int = 1) -> Categorical:
    """Shift factor levels to left or right, wrapping around at end

    Args:
        f: A factor.
        n: Positive values shift to the left; negative values shift to
            the right.

    Returns:
        The factor with levels shifted
    """
    nlvls = nlevels(_f, __calling_env=CallingEnvs.REGULAR)
    lvl_order = (seq_len(nlvls, base0_=True) + n) % nlvls

    return lvls_reorder(
        _f,
        lvl_order,
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )



"""Some default values used in datar"""
from pipda import Symbolic

# When unable to retrieve the expression for column name
# For example:
# >>> df >> mutate(f.x+1)
# In such a case, `f.x+1` cannot be fetched, we use `Var0` instead.
DEFAULT_COLUMN_PREFIX = "_Var"
NA_REPR = "<NA>"

f = Symbolic()



"""Some functions from R-base

If a function uses DataFrame/DataFrameGroupBy as first argument, it may be
registered by `register_verb` and should be placed in `./verbs.py`
"""
import itertools
from typing import Any, Callable, Iterable, List, Union
import numpy

import pandas
from pandas import Categorical, DataFrame
from pipda import register_func

from ..core.middlewares import WithDataEnv
from ..core.types import NumericType, is_scalar
from ..core.contexts import Context
from ..core.utils import arg_match
from ..core.names import repair_names


@register_func(None, context=Context.EVAL)
def cut(
    x: Iterable[NumericType],
    breaks: Any,
    labels: Iterable[Any] = None,
    include_lowest: bool = False,
    right: bool = True,
    precision: int = 2,
    ordered_result: bool = False,
) -> Categorical:
    """Divides the range of x into intervals and codes the values in x
    according to which interval they fall. The leftmost interval corresponds
    to level one, the next leftmost to level two and so on.

    Args:
        x: a numeric vector which is to be converted to a factor by cutting.
        breaks: either a numeric vector of two or more unique cut points or
            a single number (greater than or equal to 2) giving the number of
            intervals into which x is to be cut.
        labels: labels for the levels of the resulting category. By default,
            labels are constructed using "(a,b]" interval notation.
            If labels = False, simple integer codes are returned instead
            of a factor.
        include_lowest: bool, indicating if an ‘x[i]` equal to the lowest
            (or highest, for right = FALSE) ‘breaks’ value should be included.
        right: bool, indicating if the intervals should be closed on the right
            (and open on the left) or vice versa.
        precision:integer which is used when labels are not given. It determines
            the precision used in formatting the break numbers. Note, this
            argument is different from R's API, which is dig.lab.
        ordered_result: bool, should the result be an ordered categorical?

    Returns:
        A categorical object with the cuts
    """
    if labels is None:
        ordered_result = True

    return pandas.cut(
        x,
        breaks,
        labels=labels,
        include_lowest=include_lowest,
        right=right,
        precision=precision,
        ordered=ordered_result,
    )


@register_func(None, context=Context.EVAL)
def identity(x: Any) -> Any:
    """Return whatever passed in

    Expression objects are evaluated using parent context
    """
    return x


@register_func(None, context=Context.EVAL)
def expandgrid(*args: Iterable[Any], **kwargs: Iterable[Any]) -> DataFrame:
    """Expand all combinations into a dataframe. R's `expand.grid()`"""
    iters = {}
    for i, arg in enumerate(args):
        name = getattr(arg, "name", getattr(arg, "__name__", f"Var{i}"))
        iters[name] = arg
    iters.update(kwargs)

    return DataFrame(
        list(itertools.product(*iters.values())), columns=iters.keys()
    )


@register_func(None, context=Context.EVAL)
def outer(x, y, fun: Union[str, Callable] = "*") -> DataFrame:
    """Compute the outer product of two vectors.

    Args:
        x: The first vector
        y: The second vector
        fun: The function to handle how the result of the elements from
            the first and second vectors should be computed.
            The function has to be vectorized at the second argument, and
            return the same shape as y.

    Returns:
        The data frame of the outer product of x and y
    """
    if fun == "*":
        return DataFrame(numpy.outer(x, y))

    return DataFrame([fun(xelem, y) for xelem in x])


@register_func(None, context=Context.EVAL)
def make_names(names: Any, unique: bool = False) -> List[str]:
    """Make names available as columns and can be accessed by `df.<name>`

    The names will be transformed using `python-slugify` with
    `lowercase=False` and `separator="_"`. When the first character is
    a digit, preface it with "_".

    If `unique` is True, the results will be fed into
    `datar.core.names.repair_names(names, "unique")`

    Args:
        names: The names
            if it is scalar, will make it into a list.
            Then all elements will be converted into strings
        unique: Whether to make the names unique

    Returns:
        Converted names
    """
    from slugify import slugify
    from . import as_character

    if is_scalar(names):
        names = [names]
    names = as_character(names)
    names = [slugify(name, separator="_", lowercase=False) for name in names]
    names = [f"_{name}" if name[0].isdigit() else name for name in names]
    if unique:
        return repair_names(names, "unique")
    return names


@register_func(None, context=Context.EVAL)
def make_unique(names: Any) -> List[str]:
    """Make the names unique.

    It's a shortcut for `make_names(names, unique=True)`

    Args:
        names: The names
            if it is scalar, will make it into a list.
            Then all elements will be converted into strings

    Returns:
        Converted names
    """
    return make_names(names, unique=True)


@register_func(None, context=Context.EVAL)
def rank(
    x: Iterable,
    na_last: bool = True,
    ties_method: str = "average",
) -> Iterable:
    """Returns the sample ranks of the values in a vector.

    Args:
        x: A numeric vector
        na_last: for controlling the treatment of `NA`s.  If `True`, missing
            values in the data are put last; if `False`, they are put
            first; if `"keep"` they are kept  with rank `NA`.
        ties_method: a character string specifying how ties are treated
            One of `average`, `first`, `dense`, `max`, and `min`
            Note that the ties_method candidates are different than the ones
            from R, because we are using `pandas.Series.rank()`. See
            https://pandas.pydata.org/docs/reference/api/pandas.Series.rank.html

    Returns:
        A numeric rank vector of the same length as `x`
    """
    ties_method = arg_match(
        ties_method,
        "ties_method",
        ["average", "first", "dense", "max", "min"],
    )

    from ..dplyr.rank import _rank
    return _rank(x, na_last, method=ties_method)

# ---------------------------------
# Plain functions
# ---------------------------------


def data_context(data: DataFrame) -> Any:
    """Evaluate verbs, functions in the
    possibly modifying (a copy of) the original data.

    It mimic the `with` function in R, but you have to write it in a python way,
    which is using the `with` statement. And you have to use it with `as`, since
    we need the value returned by `__enter__`.

    Args:
        data: The data
        func: A function that is registered by
            `pipda.register_verb` or `pipda.register_func`.
        *args: Arguments for func
        **kwargs: Keyword arguments for func

    Returns:
        The original or modified data
    """
    return WithDataEnv(data)



"""Provides forcats verbs to manipulate factor level values"""
from typing import Any, Callable, Iterable, List, Mapping, Union

import numpy
from pandas import Categorical, DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs, functype

from ..base import (
    levels,
    match,
    nlevels,
    paste0,
    sample,
    table,
    order,
    rank,
)
from ..core.contexts import Context
from ..core.types import ForcatsRegType, ForcatsType, NumericType, is_null
from ..core.utils import get_option, logger
from ..dplyr import recode_factor, if_else
from .utils import check_factor
from .lvls import lvls_reorder, lvls_revalue
from .lvl_order import fct_relevel


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_anon(
    _f: ForcatsType,
    prefix: str = "",
    base0_: bool = None,
) -> Categorical:
    """Anonymise factor levels

    Args:
        f: A factor.
        prefix: A character prefix to insert in front of the random labels.
        base0_: Whether we start from 0?
            If not given, will use `get_option('index.base.0')`

    Returns:
        The factor with levels anonymised
    """
    _f = check_factor(_f)
    base0_ = get_option("index.base.0", base0_)
    nlvls = nlevels(_f, __calling_env=CallingEnvs.REGULAR)
    ndigits = len(str(nlvls - int(base0_)))
    lvls = paste0(
        prefix,
        [str(i + int(not base0_)).rjust(ndigits, "0") for i in range(nlvls)],
    )
    _f = lvls_revalue(
        _f,
        sample(lvls, __calling_env=CallingEnvs.REGULAR),
        __calling_env=CallingEnvs.REGULAR,
    )
    return lvls_reorder(
        _f,
        match(
            lvls,
            levels(_f, __calling_env=CallingEnvs.REGULAR),
            base0_=True,
            __calling_env=CallingEnvs.REGULAR,
        ),
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_recode(
    _f: ForcatsType,
    *args: Mapping[Any, Any],
    **kwargs: Any,
) -> Categorical:
    """Change factor levels by hand

    Args:
        _f: A factor
        *args: and
        **kwargs: A sequence of named character vectors where the name
            gives the new level, and the value gives the old level.
            Levels not otherwise mentioned will be left as is. Levels can
            be removed by naming them `NULL`.
            As `NULL/None` cannot be a name of keyword arguments, replacement
            has to be specified as a dict
            (i.e. `fct_recode(x, {NULL: "apple"})`)
            If you want to replace multiple values with the same old value,
            use a `set`/`list`/`numpy.ndarray`
            (i.e. `fct_recode(x, fruit=["apple", "banana"])`).
            This is a safe way, since `set`/`list`/`numpy.ndarray` is
            not hashable to be a level of a factor.
            Do NOT use a `tuple`, as it's hashable!

            Note that the order of the name-value is in the reverse way as
            `dplyr.recode()` and `dplyr.recode_factor()`

    Returns:
        The factor recoded with given recodings
    """
    _f = check_factor(_f)

    recodings = {}  # new => old
    for arg in args:
        if not isinstance(arg, dict):
            raise ValueError("`*args` have to be all mappings.")
        recodings.update(arg)
    recodings.update(kwargs)

    lvls = levels(_f, __calling_env=CallingEnvs.REGULAR)
    for_recode = dict(zip(lvls, lvls))  # old => new
    unknown = set()
    for key, val in recodings.items():
        if isinstance(val, (numpy.ndarray, set, list)):
            for value in val:
                if value not in lvls:
                    unknown.add(value)
                else:
                    for_recode[value] = key
        else:
            if val not in lvls:
                unknown.add(val)
            else:
                for_recode[val] = key

    if unknown:
        logger.warning("[fct_recode] Unknown levels in `_f`: %s", unknown)

    return recode_factor(_f, for_recode, __calling_env=CallingEnvs.REGULAR)


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_collapse(
    _f: ForcatsType,
    other_level: Any = None,
    **kwargs: List,
) -> Categorical:
    """Collapse factor levels into manually defined groups

    Args:
        _f: A factor
        **kwargs: The levels to collapse.
            Like `name=[old_level, old_level1, ...]`. The old levels will
            be replaced with `name`
        other_level: Replace all levels not named in `kwargs`.
            If not, don't collapse them.

    Returns:
        The factor with levels collapsed.
    """
    _f = check_factor(_f)
    levs = set(lev for sublevs in kwargs.values() for lev in sublevs)

    if other_level is not None:
        lvls = levels(_f, __calling_env=CallingEnvs.REGULAR)
        kwargs[other_level] = set(lvls) - levs

    out = fct_recode(_f, kwargs)
    if other_level in kwargs:
        return fct_relevel(
            out,
            other_level,
            after=-1,
            __calling_env=CallingEnvs.REGULAR,
        )

    return out


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_lump_min(
    _f: ForcatsType,
    min: int,
    w: Iterable[NumericType] = None,
    other_level: Any = "Other",
) -> Categorical:
    """lumps levels that appear fewer than `min` times.

    Args:
        _f: A factor
        min: Preserve levels that appear at least `min` number of times.
        w: An optional numeric vector giving weights for frequency of
            each value (not level) in f.
        other_level: Value of level used for "other" values. Always
            placed at end of levels.

    Returns:
        The factor with levels lumped.
    """
    calcs = check_calc_levels(_f, w)
    _f = calcs["_f"]

    if min < 0:
        raise ValueError("`min` must be a positive number.")

    new_levels = if_else(
        calcs["count"] >= min,
        levels(_f, __calling_env=CallingEnvs.REGULAR),
        other_level,
        __calling_env=CallingEnvs,
    )

    if other_level in new_levels:
        _f = lvls_revalue(_f, new_levels, __calling_env=CallingEnvs.REGULAR)
        return fct_relevel(_f, other_level, after=-1)

    return _f


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_lump_prop(
    _f: ForcatsType,
    prop: NumericType,
    w: Iterable[NumericType] = None,
    other_level: Any = "Other",
) -> Categorical:
    """Lumps levels that appear in fewer `prop * n` times.

    Args:
        _f: A factor
        prop: Positive `prop` lumps values which do not appear at least
            `prop` of the time. Negative `prop` lumps values that
            do not appear at most `-prop` of the time.
        w: An optional numeric vector giving weights for frequency of
            each value (not level) in f.
        other_level: Value of level used for "other" values. Always
            placed at end of levels.

    Returns:
        The factor with levels lumped.
    """
    calcs = check_calc_levels(_f, w)
    _f = calcs["_f"]

    prop_n = calcs["count"] / calcs["total"]
    if prop < 0:
        new_levels = if_else(
            prop_n <= -prop,
            levels(_f, __calling_env=CallingEnvs.REGULAR),
            other_level,
            __calling_env=CallingEnvs.REGULAR,
        )
    else:
        new_levels = if_else(
            prop_n > prop,
            levels(_f, __calling_env=CallingEnvs.REGULAR),
            other_level,
            __calling_env=CallingEnvs.REGULAR,
        )

    if prop > 0 and sum(prop_n <= prop) <= 1:
        return _f

    if other_level in new_levels:
        _f = lvls_revalue(_f, new_levels, __calling_env=CallingEnvs.REGULAR)
        return fct_relevel(_f, other_level, after=-1)

    return _f


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_lump_n(
    _f: ForcatsType,
    n: int,
    w: Iterable[NumericType] = None,
    other_level: Any = "Other",
    ties_method: str = "min",
) -> Categorical:
    """Lumps all levels except for the `n` most frequent.

    Args:
        f: A factor
        n: Positive `n` preserves the most common `n` values.
            Negative `n` preserves the least common `-n` values.
            It there are ties, you will get at least `abs(n)` values.
        w: An optional numeric vector giving weights for frequency of
            each value (not level) in f.
        other_level: Value of level used for "other" values. Always
            placed at end of levels.
        ties_method A character string specifying how ties are treated.
            One of: `average`, `first`, `dense`, `max`, and `min`.

    Returns:
        The factor with levels lumped.
    """
    calcs = check_calc_levels(_f, w)
    _f = calcs["_f"]

    if n < 0:
        rnk = rank(calcs["count"], ties_method=ties_method)
        n = -n
    else:
        rnk = rank(-calcs["count"], ties_method=ties_method)

    new_levels = if_else(
        rnk <= n,
        levels(_f, __calling_env=CallingEnvs.REGULAR),
        other_level,
        __calling_env=CallingEnvs.REGULAR,
    )

    if sum(rnk > n) <= 1:
        return _f

    if other_level in new_levels:
        _f = lvls_revalue(_f, new_levels, __calling_env=CallingEnvs.REGULAR)
        return fct_relevel(_f, other_level, after=-1)

    return _f  # pragma: no cover


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_lump_lowfreq(_f: ForcatsType, other_level: Any = "Other"):
    """lumps together the least frequent levels, ensuring
    that "other" is still the smallest level.

    Args:
        f: A factor
        other_level: Value of level used for "other" values. Always
            placed at end of levels.

    Returns:
        The factor with levels lumped.
    """
    calcs = check_calc_levels(_f)
    _f = calcs["_f"]

    new_levels = if_else(
        ~in_smallest(calcs["count"]),
        levels(_f, __calling_env=CallingEnvs.REGULAR),
        other_level,
    )

    if other_level in new_levels:
        _f = lvls_revalue(_f, new_levels, __calling_env=CallingEnvs.REGULAR)
        return fct_relevel(_f, other_level, after=-1)

    return _f


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_lump(
    _f: ForcatsType,
    n: int = None,
    prop: NumericType = None,
    w: Iterable[NumericType] = None,
    other_level: Any = "Other",
    ties_method: str = "min",
) -> Categorical:
    """Lump together factor levels into "other"

    Args:
        f: A factor
        n: Positive `n` preserves the most common `n` values.
            Negative `n` preserves the least common `-n` values.
            It there are ties, you will get at least `abs(n)` values.
        prop: Positive `prop` lumps values which do not appear at least
            `prop` of the time. Negative `prop` lumps values that
            do not appear at most `-prop` of the time.
        w: An optional numeric vector giving weights for frequency of
            each value (not level) in f.
        other_level: Value of level used for "other" values. Always
            placed at end of levels.
        ties_method A character string specifying how ties are treated.
            One of: `average`, `first`, `dense`, `max`, and `min`.

    Returns:
        The factor with levels lumped.
    """
    check_calc_levels(_f, w)

    if n is None and prop is None:
        return fct_lump_lowfreq(_f, other_level=other_level)
    if prop is None:
        return fct_lump_n(
            _f,
            n=n,
            w=w,
            other_level=other_level,
            ties_method=ties_method,
        )
    if n is None:
        return fct_lump_prop(_f, prop=prop, w=w, other_level=other_level)

    raise ValueError("Must supply only one of `n` and `prop`")


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_other(
    _f: ForcatsType,
    keep: Iterable = None,
    drop: Iterable = None,
    other_level: Any = "Other",
) -> Categorical:
    """Replace levels with "other"

    Args:
        _f: A factor
        keep: and
        drop: Pick one of `keep` and `drop`:
            - `keep` will preserve listed levels, replacing all others with
                `other_level`.
            - `drop` will replace listed levels with `other_level`, keeping all
                as is.
        other_level: Value of level used for "other" values. Always
            placed at end of levels.

    Returns:
        The factor with levels replaced.
    """
    _f = check_factor(_f)

    if (keep is None and drop is None) or (
        keep is not None and drop is not None
    ):
        raise ValueError("Must supply exactly one of `keep` and `drop`")

    lvls = levels(_f, __calling_env=CallingEnvs.REGULAR)
    if keep is not None:
        lvls[~numpy.isin(lvls, keep)] = other_level
    else:
        lvls[numpy.isin(lvls, drop)] = other_level

    _f = lvls_revalue(_f, lvls, __calling_env=CallingEnvs.REGULAR)
    return fct_relevel(
        _f,
        other_level,
        after=-1,
        __calling_env=CallingEnvs.REGULAR,
    )


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_relabel(
    _f: ForcatsType,
    _fun: Callable,
    *args: Any,
    **kwargs: Any,
) -> Categorical:
    """Automatically relabel factor levels, collapse as necessary

    Args:
        _f: A factor
        _fun: A function to be applied to each level. Must accept the old
            levels and return a character vector of the same length
            as its input.
        *args: and
        **kwargs: Addtional arguments to `_fun`

    Returns:
        The factor with levels relabeled
    """
    _f = check_factor(_f)
    old_levels = levels(_f, __calling_env=CallingEnvs.REGULAR)
    if functype(_fun) != 'plain':
        kwargs["__calling_env"] = CallingEnvs.REGULAR

    new_levels = _fun(old_levels, *args, **kwargs)
    return lvls_revalue(
        _f,
        new_levels,
        __calling_env=CallingEnvs.REGULAR
    )

# -------------
# helpers
# -------------


def check_weights(
    w: Iterable[NumericType],
    n: int = None,
) -> Iterable[NumericType]:
    """Check the weights"""
    if w is None:
        return w

    if n is None:  # pragma: no cover
        n = len(w)

    if len(w) != n:
        raise ValueError(
            f"`w` must be the same length as `f` ({n}), "
            f"not length {len(w)}."
        )

    for weight in w:
        if weight < 0 or is_null(weight):
            raise ValueError(
                f"All `w` must be non-negative and non-missing, got {weight}."
            )

    return w


def check_calc_levels(
    _f: ForcatsType,
    w: Iterable[NumericType] = None,
) -> Mapping[str, Union[ForcatsType, NumericType]]:
    """Check levels to be calculated"""
    _f = check_factor(_f)
    w = check_weights(w, len(_f))

    if w is None:
        cnt = table(_f).iloc[0, :]
        total = len(_f)
    else:
        cnt = (
            DataFrame({"w": w, "f": _f})
            .groupby("f", observed=False)
            .agg("sum")
            .iloc[:, 0]
        )
        total = sum(w)

    return {"_f": _f, "count": cnt, "total": total}


def lump_cutoff(x: Iterable[NumericType]) -> int:
    """Lump together smallest groups, ensuring that the collective
    "other" is still the smallest group. Assumes x is vector
    of counts in descending order"""
    left = sum(x)

    for i, elem in enumerate(x):
        # After group, there are this many left
        left -= elem

        if elem > left:
            return i + 1

    return len(x)  # pragma: no cover


def in_smallest(x: Iterable[NumericType]) -> Iterable[bool]:
    """Check if elements in x are the smallest of x"""
    ord_x = order(
        x,
        decreasing=True,
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )
    idx = lump_cutoff(x[ord_x])

    to_lump = numpy.arange(len(x)) >= idx
    return to_lump[order(ord_x, base0_=True, __calling_env=CallingEnvs.REGULAR)]



"""Select helpers"""
import re
import builtins
from typing import Callable, Iterable, List, Union

from pandas import DataFrame
from pipda import register_func
from pipda.utils import functype, CallingEnvs

from ..core.contexts import Context
from ..core.utils import get_option, vars_select, Array
from ..core.types import StringOrIter
from ..base import setdiff, intersect
from .group_data import group_vars


@register_func(DataFrame, context=Context.EVAL)
def where(_data: DataFrame, fn: Callable) -> List[str]:
    """Selects the variables for which a function returns True.

    Args:
        _data: The data piped in
        fn: A function that returns True or False.
            Currently it has to be `register_func/register_cfunction
            registered function purrr-like formula not supported yet.

    Returns:
        The matched columns
    """
    columns = everything(_data)
    retcols = []
    pipda_type = functype(fn)
    for col in columns:
        if pipda_type == "plain":
            conditions = fn(_data[col])
        else:
            conditions = fn(_data[col], __envdata=_data)

        if isinstance(conditions, bool):
            if conditions:
                retcols.append(col)
            else:
                # pytest-cov not detecting this line
                continue  # pragma: no cover
        elif all(conditions):
            retcols.append(col)

    return retcols


@register_func(DataFrame)
def everything(_data: DataFrame) -> List[str]:
    """Matches all columns.

    Args:
        _data: The data piped in

    Returns:
        All column names of _data
    """
    return setdiff(
        _data.columns,
        group_vars(_data, __calling_env=CallingEnvs.REGULAR),
        __calling_env=CallingEnvs.REGULAR,
    )


@register_func(context=Context.SELECT)
def last_col(
    _data: DataFrame,
    offset: int = 0,
    vars: Iterable[str] = None,
) -> str:
    """Select last variable, possibly with an offset.

    Args:
        _data: The data piped in
        offset: The offset from the end.
            Note that this is 0-based, the same as `tidyverse`'s `last_col`
        vars: A set of variable names. If not supplied, the variables are
            taken from the data columns.

    Returns:
        The variable
    """
    vars = vars or _data.columns
    return vars[-(offset + 1)]


@register_func(context=Context.SELECT)
def starts_with(
    _data: DataFrame,
    match: StringOrIter,
    ignore_case: bool = True,
    vars: Iterable[str] = None,
) -> List[str]:
    """Select columns starting with a prefix.

    Args:
        _data: The data piped in
        match: Strings. If len>1, the union of the matches is taken.
        ignore_case: If True, the default, ignores case when matching names.
        vars: A set of variable names. If not supplied, the variables are
            taken from the data columns.

    Returns:
        A list of matched vars
    """
    return _filter_columns(
        vars or _data.columns,
        match,
        ignore_case,
        lambda mat, cname: cname.startswith(mat),
    )


@register_func(context=Context.SELECT)
def ends_with(
    _data: DataFrame,
    match: str,
    ignore_case: bool = True,
    vars: Iterable[str] = None,
) -> List[str]:
    """Select columns ending with a suffix.

    Args:
        _data: The data piped in
        match: Strings. If len>1, the union of the matches is taken.
        ignore_case: If True, the default, ignores case when matching names.
        vars: A set of variable names. If not supplied, the variables are
            taken from the data columns.

    Returns:
        A list of matched vars
    """
    return _filter_columns(
        vars or _data.columns,
        match,
        ignore_case,
        lambda mat, cname: cname.endswith(mat),
    )


@register_func(context=Context.SELECT)
def contains(
    _data: DataFrame,
    match: str,
    ignore_case: bool = True,
    vars: Iterable[str] = None,
) -> List[str]:
    """Select columns containing substrings.

    Args:
        _data: The data piped in
        match: Strings. If len>1, the union of the matches is taken.
        ignore_case: If True, the default, ignores case when matching names.
        vars: A set of variable names. If not supplied, the variables are
            taken from the data columns.

    Returns:
        A list of matched vars
    """
    return _filter_columns(
        vars or _data.columns,
        match,
        ignore_case,
        lambda mat, cname: mat in cname,
    )


@register_func(context=Context.SELECT)
def matches(
    _data: DataFrame,
    match: str,
    ignore_case: bool = True,
    vars: Iterable[str] = None,
) -> List[str]:
    """Select columns matching regular expressions.

    Args:
        _data: The data piped in
        match: Regular expressions. If len>1, the union of the matches is taken.
        ignore_case: If True, the default, ignores case when matching names.
        vars: A set of variable names. If not supplied, the variables are
            taken from the data columns.

    Returns:
        A list of matched vars
    """
    return _filter_columns(
        vars or _data.columns,
        match,
        ignore_case,
        re.search,
    )


@register_func(context=Context.EVAL)
def all_of(
    _data: DataFrame,
    x: Iterable[Union[int, str]],
    base0_: bool = None,
) -> List[str]:
    """For strict selection.

    If any of the variables in the character vector is missing,
    an error is thrown.

    Args:
        _data: The data piped in
        x: A set of variables to match the columns
        base0_: Whether `x` is 0-based or not.
            if not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The matched column names

    Raises:
        ColumnNotExistingError: When any of the elements in `x` does not exist
            in `_data` columns
    """
    all_columns = _data.columns
    x = all_columns[vars_select(all_columns, x, base0=base0_)]
    # where do errors raise?

    # nonexists = setdiff(x, all_columns)
    # if nonexists:
    #     raise ColumnNotExistingError(
    #         "Can't subset columns that don't exist. "
    #         f"Columns {nonexists} not exist."
    #     )

    return list(x)


@register_func(context=Context.SELECT)
def any_of(
    _data: DataFrame,
    x: Iterable[Union[int, str]],

    vars: Iterable[str] = None,
    base0_: bool = None,
) -> List[str]:
    """Select but doesn't check for missing variables.

    It is especially useful with negative selections,
    when you would like to make sure a variable is removed.

    Args:
        _data: The data piped in
        x: A set of variables to match the columns
        base0_: Whether `x` is 0-based or not.
            if not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The matched column names
    """
    vars = vars or _data.columns
    x = vars_select(vars, x, raise_nonexists=False, base0=base0_)
    # exists = []
    # for idx in x:
    #     try:
    #         exists.append(vars[idx])
    #     except IndexError:
    #         ...
    # do we need intersect?
    return intersect(
        vars, [vars[idx] for idx in x], __calling_env=CallingEnvs.REGULAR
    )


@register_func(None)
def num_range(
    prefix: str,
    range: Iterable[int],
    width: int = None,
    base0_: bool = None,
) -> List[str]:
    """Matches a numerical range like x01, x02, x03.

    Args:
        _data: The data piped in
        prefix: A prefix that starts the numeric range.
        range_: A sequence of integers, like `range(3)` (produces `0,1,2`).
        width: Optionally, the "width" of the numeric range.
            For example, a range of 2 gives "01", a range of three "001", etc.
        base0_: Whether it is 0-based

    Returns:
        A list of ranges with prefix.
    """
    base0_ = get_option("index.base.0", base0_)
    zfill = lambda elem: (
        elem + int(not base0_)
        if not width
        else str(elem + int(not base0_)).zfill(width)
    )
    return Array([f"{prefix}{zfill(elem)}" for elem in builtins.range(range)])


def _filter_columns(
    all_columns: Iterable[str],
    match: Union[Iterable[str], str],
    ignore_case: bool,
    func: Callable[[str, str], bool],
) -> List[str]:
    """Filter the columns with given critera

    Args:
        all_columns: The column pool to filter
        match: Strings. If len>1, the union of the matches is taken.
        ignore_case: If True, the default, ignores case when matching names.
        func: A function to define how to filter.

    Returns:
        A list of matched vars
    """
    if not isinstance(match, (tuple, list, set)):
        match = [match]  # type: ignore

    ret = []
    for mat in match:
        for column in all_columns:
            if column in ret:
                continue
            if func(
                mat.lower() if ignore_case else mat,
                column.lower() if ignore_case else column,
            ):
                ret.append(column)

    return ret



"""Generating and manipulating sequences"""

from typing import Any, Iterable, Union

import numpy
import pandas
from pipda import register_func

from ..core.collections import Collection
from ..core.contexts import Context
from ..core.types import (
    ArrayLikeType,
    IntOrIter,
    IntType,
    NumericType,
    is_iterable,
    is_not_null,
    is_scalar,
    is_scalar_int,
    is_null,
)
from ..core.utils import Array, get_option, length_of, logger


@register_func(None, context=Context.EVAL)
def seq_along(along_with: Iterable[Any], base0_: bool = None) -> ArrayLikeType:
    """Generate sequences along an iterable

    Args:
        along_with: An iterable to seq along with
        base0_: Whether the generated sequence should be 0-based.
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The generated sequence.
    """
    base0_ = get_option("index.base.0", base0_)
    return Array(range(len(along_with))) + int(not base0_)


@register_func(None, context=Context.EVAL)
def seq_len(length_out: IntOrIter, base0_: bool = None) -> ArrayLikeType:
    """Generate sequences with the length"""
    base0_ = get_option("index.base.0", base0_)
    if is_scalar(length_out):
        return Array(range(int(length_out))) + int(not base0_)
    if len(length_out) > 1:
        logger.warning(
            "In seq_len(%r) : first element used of 'length_out' argument",
            length_out,
        )
    length_out = int(list(length_out)[0])
    return Array(range(length_out)) + int(not base0_)


@register_func(None, context=Context.EVAL)
def seq(
    from_: IntType = None,
    to: IntType = None,
    by: IntType = None,
    length_out: IntType = None,
    along_with: IntType = None,
    base0_: bool = None,
) -> ArrayLikeType:
    """Generate a sequence

    https://rdrr.io/r/base/seq.html

    Note that this API is consistent with r-base's seq. 1-based and inclusive.
    """
    base0_ = get_option("index.base.0", base0_)
    if along_with is not None:
        return seq_along(along_with, base0_)
    if from_ is not None and not is_scalar(from_):
        return seq_along(from_, base0_)
    if length_out is not None and from_ is None and to is None:
        return seq_len(length_out)

    base = int(not base0_)

    if from_ is None:
        from_ = base
    elif to is None:
        from_, to = base, from_

    if length_out is not None:
        by = (float(to) - float(from_)) / float(length_out - 1)

    elif by is None:
        by = 1 if to > from_ else -1
        length_out = to - from_ + base if to > from_ else from_ - to + base
    else:
        length_out = (to - from_ + 0.1 * by + float(base) * by) // by
    return Array([from_ + n * by for n in range(int(length_out))])


@register_func(None, context=Context.EVAL)
def rep(
    x: Any,
    times: IntOrIter = 1,
    length: int = None,
    each: int = 1,
) -> ArrayLikeType:
    """replicates the values in x

    Args:
        x: a vector or scaler
        times: number of times to repeat each element if of length len(x),
            or to repeat the whole vector if of length 1
        length: non-negative integer. The desired length of the output vector
        each: non-negative integer. Each element of x is repeated each times.

    Returns:
        A list of repeated elements in x.
    """
    if is_scalar(x):
        x = [x]
    if is_iterable(times):
        if len(times) != len(x):
            raise ValueError(
                "Invalid times argument, expect length "
                f"{len(times)}, got {len(x)}"
            )
        if each != 1:
            raise ValueError(
                "Unexpected each argument when times is an iterable."
            )

    if is_scalar_int(times):
        x = Array([elem for elem in x for _ in range(each)] * int(times))
    else:
        x = Array([elem for n, elem in zip(times, x) for _ in range(n)])
    if length is None:
        return x
    repeats = length // len(x) + 1
    x = numpy.tile(x, repeats)
    return x[:length]


@register_func(None, context=Context.EVAL)
def rev(x: Iterable[Any]) -> numpy.ndarray:
    """Get reversed vector"""
    dtype = getattr(x, "dtype", None)
    return Array(list(reversed(x)), dtype=dtype)


@register_func(None, context=Context.EVAL)
def unique(x: Iterable[Any]) -> numpy.ndarray:
    """Get unique elements"""
    # return numpy.unique(x)
    return pandas.unique(x)  # keeps order


@register_func(None, context=Context.EVAL)
def sort(
    x: Iterable[Any],
    decreasing: bool = False,
    na_last: bool = None,
) -> numpy.ndarray:
    """Sorting or Ordering Vectors

    Args:
        x: A vector to be sorted
        decreasing: Should the vector sort be increasing or decreasing?
        na_last: for controlling the treatment of `NA`s.  If `True`, missing
            values in the data are put last; if `FALSE`, they are put
            first; if `NA`, they are removed.

    Returns:
        The sorted array
    """
    x = Array(x)
    not_na_idx = is_not_null(x)
    non_nas = x[not_na_idx]
    out = numpy.sort(non_nas)
    if decreasing:
        out = out[::-1]
    if na_last is None:
        return out

    nas = x[~not_na_idx].tolist()
    if na_last:
        return Array(out.tolist() + nas)
    return Array(nas + out.tolist())


@register_func(None, context=Context.EVAL)
def order(
    x: Iterable[Any],
    na_last: bool = None,
    decreasing: bool = False,
    # method?
    base0_: bool = None,
) -> numpy.ndarray:
    """Returns a permutation which rearranges its first argument
    into ascending or descending order

    Args:
        x: A vector to order
        decreasing: Should the vector sort be increasing or decreasing?
        na_last: for controlling the treatment of `NA`s.  If `True`, missing
            values in the data are put last; if `FALSE`, they are put
            first; if `NA`, they are removed.
        base0_: Return a 0-based indices or not.
            If not given, will use `get_option('which.base.0')`

    Returns:
        The ordered indices of the vector
    """
    na_mask = is_null(x)
    na_idx = numpy.flatnonzero(na_mask).tolist()
    out_idx = numpy.argsort(x)
    if decreasing:
        out_idx = out_idx[::-1]

    non_na_idx = [idx for idx in out_idx if idx not in na_idx]
    if na_last is None:
        out = non_na_idx
    elif na_last:
        out = non_na_idx + na_idx
    else:
        out = na_idx + non_na_idx

    out = Array(out, dtype=int)
    return out if base0_ else out + 1


length = register_func(None, context=Context.EVAL, func=length_of)


@register_func(None, context=Context.EVAL)
def lengths(x: Any) -> IntOrIter:
    """Lengths of elements in x"""
    if is_scalar(x):
        return Array([1], dtype=numpy.int_)
    return Array([length(elem) for elem in x], dtype=numpy.int_)


@register_func(None, context=Context.EVAL)
def sample(
    x: Union[IntType, Iterable[Any]],
    size: int = None,
    replace: bool = False,
    prob: Iterable[NumericType] = None,
) -> Iterable[Any]:
    """Takes a sample of the specified size from the elements of x using
    either with or without replacement.

    https://rdrr.io/r/base/sample.html

    Args:
        x: either a vector of one or more elements from which to choose,
            or a positive integer.
        n: a positive number, the number of items to choose from.
        size: a non-negative integer giving the number of items to choose.
        replace: should sampling be with replacement?
        prob: a vector of probability weights for obtaining the elements of
            the vector being sampled.

    Returns:
        A vector of length size with elements drawn from either x or from the
        integers 1:x.
    """
    if isinstance(x, str):
        x = list(x)
    if size is None:
        size = len(x) if is_iterable(x) else x
    return numpy.random.choice(x, int(size), replace=replace, p=prob)


@register_func(None, context=Context.UNSET)
def c(*elems: Any) -> Collection:
    """Mimic R's concatenation. Named one is not supported yet
    All elements passed in will be flattened.

    Args:
        *elems: The elements

    Returns:
        A collection of elements
    """
    return Collection(*elems)


@register_func(None, context=Context.EVAL)
def match(
    x: Any,
    table: Iterable,
    nomatch: Any = -1,
    # incomparables ...,
    base0_: bool = None,
) -> Iterable[int]:
    """match returns a vector of the positions of (first) matches of
    its first argument in its second.

    See stackoverflow#4110059/pythonor-numpy-equivalent-of-match-in-r

    Args:
        x: The values to be matched
        table: The values to be matched against
        nomatch: The value to be returned in the case when no match is found
            Instead of NA in R, this function takes -1 for non-matched elements
            to keep the type as int.
        base0_: Whether return indices are 0-based or not.
            If not provided, will use `get_option('which_base_0')`
    """
    base = int(not get_option("which_base_0", base0_))
    return Array(
        [
            list(table).index(elem) + base if elem in table else nomatch
            for elem in x
        ],
        dtype=int,
    )



"""Trigonometric and Hyperbolic Functions"""

from typing import Callable

import numpy
from pipda import register_func

from ..core.contexts import Context
from ..core.types import FloatOrIter
from .constants import pi


def _register_trig_hb_func(name: str, np_name: str, doc: str) -> Callable:
    """Register trigonometric and hyperbolic function"""
    np_fun = getattr(numpy, np_name)
    if name.endswith("pi"):
        func = lambda x: np_fun(x * pi)
    else:
        # ufunc cannot set context
        func = lambda x: np_fun(x)

    func = register_func(None, context=Context.EVAL, func=func)
    func.__name__ = name
    func.__doc__ = doc
    return func


sin = _register_trig_hb_func(
    "sin",
    "sin",
    doc="""The sine function

Args:
    x: a numeric value or iterable

Returns:
    The sine value of `x`
""",
)

cos = _register_trig_hb_func(
    "cos",
    "cos",
    doc="""The cosine function

Args:
    x: a numeric value or iterable

Returns:
    The cosine value of `x`
""",
)

tan = _register_trig_hb_func(
    "tan",
    "tan",
    doc="""The tangent function

Args:
    x: a numeric value or iterable

Returns:
    The tangent value of `x`
""",
)

acos = _register_trig_hb_func(
    "acos",
    "arccos",
    doc="""The arc-cosine function

Args:
    x: a numeric value or iterable

Returns:
    The arc-cosine value of `x`
""",
)

asin = _register_trig_hb_func(
    "acos",
    "arcsin",
    doc="""The arc-sine function

Args:
    x: a numeric value or iterable

Returns:
    The arc-sine value of `x`
""",
)

atan = _register_trig_hb_func(
    "acos",
    "arctan",
    doc="""The arc-sine function

Args:
    x: a numeric value or iterable

Returns:
    The arc-sine value of `x`
""",
)

sinpi = _register_trig_hb_func(
    "sinpi",
    "sin",
    doc="""The sine function

Args:
    x: a numeric value or iterable, which is the multiple of pi

Returns:
    The sine value of `x`
""",
)

cospi = _register_trig_hb_func(
    "cospi",
    "cos",
    doc="""The cosine function

Args:
    x: a numeric value or iterable, which is the multiple of pi

Returns:
    The cosine value of `x`
""",
)

tanpi = _register_trig_hb_func(
    "tanpi",
    "tan",
    doc="""The tangent function

Args:
    x: a numeric value or iterable, which is the multiple of pi

Returns:
    The tangent value of `x`
""",
)

cosh = _register_trig_hb_func(
    "cosh",
    "cosh",
    doc="""Hyperbolic cosine

Args:
    x: a numeric value or iterable

Returns:
    The hyperbolic cosine value of `x`
""",
)

sinh = _register_trig_hb_func(
    "sinh",
    "sinh",
    doc="""Hyperbolic sine

Args:
    x: a numeric value or iterable

Returns:
    The hyperbolic sine value of `x`
""",
)

tanh = _register_trig_hb_func(
    "tanh",
    "tanh",
    doc="""Hyperbolic tangent

Args:
    x: a numeric value or iterable

Returns:
    The hyperbolic tangent value of `x`
""",
)

acosh = _register_trig_hb_func(
    "acosh",
    "arccosh",
    doc="""Hyperbolic arc-cosine

Args:
    x: a numeric value or iterable

Returns:
    The hyperbolic arc-cosine value of `x`
""",
)

asinh = _register_trig_hb_func(
    "asinh",
    "arcsinh",
    doc="""Hyperbolic arc-sine

Args:
    x: a numeric value or iterable

Returns:
    The hyperbolic arc-sine value of `x`
""",
)

atanh = _register_trig_hb_func(
    "atanh",
    "arctanh",
    doc="""Hyperbolic arc-tangent

Args:
    x: a numeric value or iterable

Returns:
    The hyperbolic arc-tangent value of `x`
""",
)


@register_func(None, context=Context.EVAL)
def atan2(y: FloatOrIter, x: FloatOrIter) -> FloatOrIter:
    """Calculates the angle between the x-axis and the vector (0,0) -> (x,y)

    Args:
        y: and
        x: The end coordinates of the vector

    Returns:
        The angle between x-axis and vector (0,0) -> (x,y)
    """
    return numpy.arctan2(y, x)



"""Some functions ported from R-stats"""
from typing import Any, Iterable, List

import numpy
from pandas import Series
from pipda import register_func

from ..core.types import (
    FloatOrIter,
    NumericOrIter,
    NumericType,
    ArrayLikeType,
    is_scalar,
)
from ..core.contexts import Context
from ..core.utils import Array
from ..base import NA


@register_func(None, context=Context.EVAL)
def rnorm(n: int, mean: float = 0.0, sd: float = 1.0) -> List[float]:
    """random generation for the normal distribution with mean equal to mean
    and standard deviation equal to sd.

    Args:
        n: number of observations. No iterables allowed
        mean: means. No iterables allowed
        sd: standard deviations. No iterables allowed


    Returns:
        Randomly generated deviates.
    """
    return numpy.random.normal(loc=mean, scale=sd, size=n)


def runif(n: int, min: float = 0.0, max: float = 1.0) -> List[float]:
    """random generation for the uniform distribution

    Args:
        n: number of observations. No iterables allowed
        min: the minima. No iterables allowed
        max: the maxima. No iterables allowed


    Returns:
        Randomly generated deviates.
    """
    return numpy.random.uniform(low=min, high=max, size=n)


def rpois(n: int, lambda_: float) -> List[float]:
    """random generation for the Poisson distribution with parameter lambda_.

    Args:
        n: number of random values to return.
        mean: non-negative means.


    Returns:
        Randomly generated deviates.
    """
    return numpy.random.poisson(lam=lambda_, size=n)


@register_func(None, context=Context.EVAL)
def quantile(
    series: Iterable[Any],
    probs: FloatOrIter = (0.0, 0.25, 0.5, 0.75, 1.0),
    na_rm: bool = False,
) -> ArrayLikeType:
    """produces sample quantiles corresponding to the given probabilities.

    Args:
        series: The data to sample
        probs: numeric vector of probabilities with values in [0,1]
        na_rm: if true, any ‘NA’ and ‘NaN’'s are removed from ‘x’
            before the quantiles are computed.

    Returns:
        An array of quantile values
    """
    return (
        numpy.nanquantile(series, probs)
        if na_rm
        else numpy.quantile(series, probs)
    )


@register_func(None, context=Context.EVAL)
def std(
    x: Iterable[Any],
    na_rm: bool = False,
    # numpy default is 0. Make it 1 to be consistent with R
    ddof: int = 1,
) -> float:
    """Get standard deviation of the input"""
    if isinstance(x, Series):
        return x.std(skipna=na_rm, ddof=ddof)

    return numpy.nanstd(x, ddof=ddof) if na_rm else numpy.std(x, ddof=ddof)


sd = std


@register_func(None, context=Context.EVAL)
def weighted_mean(
    x: NumericOrIter,
    w: NumericOrIter = None,
    na_rm: bool = False,
) -> NumericType:
    """Calculate weighted mean"""
    if is_scalar(x):
        x = [x]  # type: ignore
    if w is not None and is_scalar(w):
        w = [w]  # type: ignore
    x = Array(x)
    if w is not None:
        w = Array(w)
        if len(x) != len(w):
            raise ValueError("'x' and 'w' must have the same length")

    if na_rm:
        notna = ~numpy.isnan(x)
        x = x[notna]
        if w is not None:
            w = w[notna]

    if w is not None and sum(w) == 0:
        return NA
    return numpy.average(x, weights=w)



"""Verbs ported from R-utils"""
from typing import Sequence, Union

import numpy
from pandas import DataFrame, Series

from pipda import register_verb


@register_verb((DataFrame, Series, list, tuple, numpy.ndarray))
def head(
    _data: Union[DataFrame, Series, Sequence, numpy.ndarray], n: int = 6
) -> DataFrame:
    """Get the first n rows of the dataframe or a vector

    Args:
        n: The number of rows/elements to return

    Returns:
        The dataframe with first n rows or a vector with first n elements
    """
    if isinstance(_data, DataFrame):
        return _data.head(n)
    return _data[:n]


@register_verb((DataFrame, Series, list, tuple, numpy.ndarray))
def tail(
    _data: Union[DataFrame, Series, Sequence, numpy.ndarray], n: int = 6
) -> DataFrame:
    """Get the last n rows of the dataframe or a vector

    Args:
        n: The number of rows/elements to return

    Returns:
        The dataframe with last n rows or a vector with last n elements
    """
    if isinstance(_data, DataFrame):
        return _data.tail(n)

    out = _data[-n:]
    try:
        return out.reset_index(drop=True)  # type: ignore[union-attr]
    except AttributeError:
        return out



"""Types for convenience"""
from typing import Any, Iterable, Union, List, Type, Tuple

import numpy

from pandas import Categorical, Index, Series, isnull as is_null
from pandas.api.types import is_scalar as is_scalar_, is_list_like
from pandas.core.dtypes.common import is_categorical_dtype as is_categorical

Dtype = Union[
    str, numpy.dtype, Type[Union[str, float, int, complex, bool, object]]
]
NumericType = Union[int, float, complex, numpy.number]
IntType = Union[int, numpy.integer]
FloatType = Union[float, numpy.float_]
ComplexType = Union[complex, numpy.complex_]
StringType = Union[str, numpy.str_]
ArrayLikeType = Union[Series, List, Tuple, numpy.ndarray, Index]
CategoricalLikeType = Union[Series, Categorical]
StringOrIter = Union[StringType, Iterable[StringType]]
IntOrIter = Union[IntType, Iterable[IntType]]
DoubleOrIter = Union[numpy.double, Iterable[numpy.double]]
BoolOrIter = Union[bool, Iterable[bool]]
FloatOrIter = Union[FloatType, Iterable[FloatType]]
NumericOrIter = Union[NumericType, Iterable[NumericType]]
ComplexOrIter = Union[ComplexType, Iterable[ComplexType]]
TypeOrIter = Union[Type, Iterable[Type]]
DtypeOrIter = Union[Dtype, Iterable[Dtype]]
NoneType = type(None)
# nan is float, a better NA annotation?
NAType = Union[NoneType, type(numpy.nan)]   # type: ignore
NAOrIter = Union[NAType, Iterable[NAType]]  # type: ignore

# Used for register_verb in forcats
# Except for cattegorical, string vectors are also avaiable
ForcatsRegType = (Series, Categorical, list, tuple, Index, numpy.ndarray)
ForcatsType = Union[CategoricalLikeType, ArrayLikeType]


# used for type checks
def is_scalar_int(x: Any) -> bool:
    """Check if a value is an integer"""
    return isinstance(x, (int, numpy.integer))


def is_scalar(x: Any) -> bool:
    """Check if a value is scalar.

    None will be counted as scalar
    """
    ret = is_scalar_(x)
    if ret:
        return ret
    try:
        iter(x)
    except TypeError:
        return True
    return False


def is_iterable(x: Any) -> bool:
    """Check if a value is iterable, which is not a scalar"""
    return not is_scalar(x)


def is_not_null(x: Any) -> BoolOrIter:
    """Invert of is_null (pandas.isnull)"""
    return ~is_null(x)



"""Basic functions"""
from typing import Any, Iterable
import numpy

from pandas import Series
from pipda import register_func

from ..core.contexts import Context
from ..core.collections import Collection
from ..core.utils import Array


@register_func(None, context=Context.EVAL)
def itemgetter(data: Iterable[Any], *subscripts) -> numpy.ndarray:
    """Itemgetter as a function for verb

    In datar expression, we can do:
    >>> arr = [1,2,3]
    >>> tibble(x=2) >> mutate(y=arr[f.x])

    Since `arr[f.x]` won't compile. We need to use the `itemgetter` operator:
    >>> tibble(x=2) >> mutate(y=itemgetter(arr, f.x))

    Args:
        data: The data to be get items from
        *subscripts: The subscripts
    """
    flattened = Collection(
        subs.values if isinstance(subs, Series) else subs for subs in subscripts
    )

    return Array([data[sub] for sub in flattened])



"""Provide options"""

from typing import Any, Generator, Mapping, Union, Callable
from contextlib import contextmanager

from diot import Diot


_key_transform = lambda key: key.replace("_", ".")
_dict_transform_back = lambda dic: {
    key.replace(".", "_"): val for key, val in dic.items()
}

OPTIONS = Diot(
    # Whether use 0-based numbers when index is involved, acts similar like R
    # Otherwise, like python
    index_base_0=False,
    # Whether which, which_min, which_max is 0-based
    which_base_0=True,
    dplyr_summarise_inform=True,
    # whether warn about importing functions that override builtin ones.
    warn_builtin_names=True,
    add_option=True,
    # allow 'a.b' to access 'a_b'
    diot_transform=_key_transform,
)

OPTION_CALLBACKS = Diot(
    # allow 'a.b' to access 'a_b'
    diot_transform=_key_transform
)


def options(
    *args: Union[str, Mapping[str, Any]],
    **kwargs: Any,
) -> Mapping[str, Any]:
    """Allow the user to set and examine a variety of global options

    Args:
        *args: Names of options to return
        **kwargs: name-value pair to create/set an option
    """
    if not args and not kwargs:
        return OPTIONS.copy()

    names = [arg.replace(".", "_") for arg in args if isinstance(arg, str)]
    pairs = {}
    for arg in args:
        if isinstance(arg, dict):
            pairs.update(_dict_transform_back(arg))
    pairs.update(_dict_transform_back(kwargs))

    out = Diot(
        {
            name: value
            for name, value in OPTIONS.items()
            if name in names or name in pairs
        },
        diot_transform=_key_transform,
    )

    for key, val in pairs.items():
        oldval = OPTIONS[key]
        if oldval == val:
            continue
        OPTIONS[key] = val
        callback = OPTION_CALLBACKS.get(key)
        if callable(callback):
            callback(val)

    return out


@contextmanager
def options_context(**kwargs: Any) -> Generator:
    """A context manager to execute code with temporary options

    Note that this is not thread-safe.
    """
    opts = options()  # type: Mapping[str, Any]
    options(**kwargs)
    yield
    options(opts)


def get_option(x: str, default: Any = None) -> Any:
    """Get the current value set for option ‘x’,
    or ‘default’ (which defaults to ‘NULL’) if the option is unset.

    Args:
        x: The name of the option
        default: The default value if `x` is unset
    """
    return OPTIONS.get(x, default)


def add_option(x: str, default: Any = None, callback: Callable = None) -> None:
    """Add an option"""
    OPTIONS[x] = default
    if callback:
        OPTION_CALLBACKS[x] = callback
        callback(default)



"""Rename columns

https://github.com/tidyverse/dplyr/blob/master/R/rename.R
"""
from typing import Any, Callable

from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.utils import vars_select
from ..core.grouped import DataFrameGroupBy
from .group_data import group_vars
from .select import _eval_select


@register_verb(DataFrame, context=Context.SELECT)
def rename(_data: DataFrame, base0_: bool = None, **kwargs: str) -> DataFrame:
    """Changes the names of individual variables using new_name = old_name
    syntax

    Args:
        _data: The dataframe
        **kwargs: The new_name = old_name pairs
        base0_: Whether the old_name is 0-based if given by indexes.
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The dataframe with new names
    """
    gvars = group_vars(_data, __calling_env=CallingEnvs.REGULAR)
    all_columns = _data.columns
    selected, new_names = _eval_select(
        all_columns,
        _group_vars=gvars,
        base0_=base0_,
        **kwargs,
    )

    if isinstance(_data, DataFrameGroupBy):
        out = _data.copy(copy_grouped=True)
    else:
        out = _data.copy()
    # new_names: old -> new
    # cannot do with duplicates
    # out.rename(columns=new_names, inplace=True)
    out.columns = [
        new_names.get(col, col) if i in selected else col
        for i, col in enumerate(all_columns)
    ]

    if isinstance(out, DataFrameGroupBy):
        out.attrs["_group_vars"] = [new_names.get(name, name) for name in gvars]
        out._group_data.columns = out.attrs["_group_vars"] + ["_rows"]
    # attrs copied?
    return out


@register_verb(DataFrame, context=Context.SELECT)
def rename_with(
    _data: DataFrame,
    _fn: Callable[[str], str],
    *args: Any,
    base0_: bool = None,
    **kwargs: Any,
) -> DataFrame:
    """Renames columns using a function.

    Args:
        _data: The dataframe
        _fn: The function to rename a column
        *args: the columns to rename and non-keyword arguments for the `_fn`.
            If `*args` is not provided, then assuming all columns, and
            no non-keyword arguments are allowed to pass to the function, use
            keyword arguments instead.
        **kwargs: keyword arguments for `_fn`
        base0_: Whether the old_name is 0-based if given by indexes.
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The dataframe with new names
    """
    if not args:
        cols = _data.columns.tolist()
    else:
        cols = args[0]
        args = args[1:]

    cols = _data.columns[vars_select(_data.columns, cols, base0=base0_)]
    new_columns = {
        _fn(col, *args, **kwargs): col for col in cols  # type: ignore
    }
    return rename(
        _data, **new_columns, base0_=True, __calling_env=CallingEnvs.REGULAR
    )



"""Port of forcats"""
from .lvl_order import (
    as_factor,
    fct_relevel,
    fct_inorder,
    fct_infreq,
    fct_inseq,
    fct_reorder,
    fct_reorder2,
    fct_rev,
    fct_shift,
    fct_shuffle,
    first2,
    last2,
)
from .lvl_value import (
    fct_anon,
    fct_collapse,
    fct_lump,
    fct_lump_lowfreq,
    fct_lump_min,
    fct_lump_n,
    fct_lump_prop,
    fct_other,
    fct_recode,
    fct_relabel,
)
from .lvl_addrm import (
    fct_drop,
    fct_expand,
    fct_explicit_na,
    fct_unify,
)
from .fct_multi import fct_c, fct_cross
from .misc import (
    fct_count,
    fct_match,
    fct_unique,
)
from .lvls import (
    lvls_expand,
    lvls_reorder,
    lvls_revalue,
    lvls_union,
)



"""Mutating joins"""
from typing import Iterable, Mapping, Union

import pandas
from pandas import DataFrame, Series, Categorical
from pandas.core.dtypes.common import is_categorical_dtype
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.types import StringOrIter, is_scalar
from ..core.grouped import DataFrameGroupBy
from ..core.utils import reconstruct_tibble
from ..base import intersect, setdiff, union
from .group_by import group_by_drop_default
from .group_data import group_data, group_vars
from .dfilter import filter as filter_


def _join(
    x: DataFrame,
    y: DataFrame,
    how: str,
    by: Union[StringOrIter, Mapping[str, str]] = None,
    copy: bool = False,
    suffix: Iterable[str] = ("_x", "_y"),
    # na_matches: str = "", # TODO: how?
    keep: bool = False,
) -> DataFrame:
    """General join"""
    if by is not None and not by:
        ret = pandas.merge(x, y, how="cross", copy=copy, suffixes=suffix)
    elif isinstance(by, dict):
        right_on = list(by.values())
        ret = pandas.merge(
            x,
            y,
            left_on=list(by.keys()),
            right_on=right_on,
            how=how,
            copy=copy,
            suffixes=suffix,
        )
        if not keep:
            ret.drop(columns=right_on, inplace=True)
    elif keep:
        if by is None:
            by = intersect(
                x.columns, y.columns, __calling_env=CallingEnvs.REGULAR
            )
        # on=... doesn't keep both by columns in left and right
        left_on = [f"{col}{suffix[0]}" for col in by]
        right_on = [f"{col}{suffix[1]}" for col in by]
        x = x.rename(columns=dict(zip(by, left_on)))
        y = y.rename(columns=dict(zip(by, right_on)))
        ret = pandas.merge(
            x,
            y,
            left_on=left_on,
            right_on=right_on,
            how=how,
            copy=copy,
            suffixes=suffix,
        )
    else:
        if by is None:
            by = intersect(
                x.columns, y.columns, __calling_env=CallingEnvs.REGULAR
            )
        by = [by] if is_scalar(by) else by  # type: ignore
        ret = pandas.merge(x, y, on=by, how=how, copy=copy, suffixes=suffix)
        for col in by:
            if is_categorical_dtype(x[col]) and is_categorical_dtype(y[col]):
                ret[col] = Categorical(
                    ret[col],
                    categories=union(
                        x[col].cat.categories,
                        y[col].cat.categories,
                        __calling_env=CallingEnvs.REGULAR,
                    ),
                )

    return reconstruct_tibble(x, ret, keep_rowwise=True)


@register_verb(
    DataFrame,
    context=Context.EVAL,
    extra_contexts={"by": Context.SELECT},
)
def inner_join(
    x: DataFrame,
    y: DataFrame,
    by: Union[StringOrIter, Mapping[str, str]] = None,
    copy: bool = False,
    suffix: Iterable[str] = ("_x", "_y"),
    keep: bool = False,
) -> DataFrame:
    """Mutating joins including all rows in x and y.

    Args:
        x, y: A pair of data frames
        by: A character vector of variables to join by.
            If keys from `x` and `y` are different, use a dict
            (i.e. `{"colX": "colY"}`) instead of a list.
        copy: If x and y are not from the same data source, and copy is
            TRUE, then y will be copied into the same src as x.
            This allows you to join tables across srcs, but it is a
            potentially expensive operation so you must opt into it.
        suffix: If there are non-joined duplicate variables in x and y,
            these suffixes will be added to the output to disambiguate them.
            Should be a character vector of length 2.
        keep: Should the join keys from both x and y be preserved in the output?

    Returns:
        The joined dataframe
    """
    return _join(
        x,
        y,
        how="inner",
        by=by,
        copy=copy,
        suffix=suffix,
        keep=keep,
    )


@register_verb(
    DataFrame,
    context=Context.EVAL,
    extra_contexts={"by": Context.SELECT},
)
def left_join(
    x: DataFrame,
    y: DataFrame,
    by: Union[StringOrIter, Mapping[str, str]] = None,
    copy: bool = False,
    suffix: Iterable[str] = ("_x", "_y"),
    keep: bool = False,
) -> DataFrame:
    """Mutating joins including all rows in x.

    See Also:
        [`inner_join()`](datar.dplyr.join.inner_join)
    """
    return _join(
        x,
        y,
        how="left",
        by=by,
        copy=copy,
        suffix=suffix,
        keep=keep,
    )


@register_verb(
    DataFrame,
    context=Context.EVAL,
    extra_contexts={"by": Context.SELECT},
)
def right_join(
    x: DataFrame,
    y: DataFrame,
    by: Union[StringOrIter, Mapping[str, str]] = None,
    copy: bool = False,
    suffix: Iterable[str] = ("_x", "_y"),
    keep: bool = False,
) -> DataFrame:
    """Mutating joins including all rows in y.

    See Also:
        [`inner_join()`](datar.dplyr.join.inner_join)

    Note:
        The rows of the order is preserved according to `y`. But `dplyr`'s
        `right_join` preserves order from `x`.
    """
    return _join(
        x,
        y,
        how="right",
        by=by,
        copy=copy,
        suffix=suffix,
        keep=keep,
    )


@register_verb(
    DataFrame,
    context=Context.EVAL,
    extra_contexts={"by": Context.SELECT},
)
def full_join(
    x: DataFrame,
    y: DataFrame,
    by: Union[StringOrIter, Mapping[str, str]] = None,
    copy: bool = False,
    suffix: Iterable[str] = ("_x", "_y"),
    keep: bool = False,
) -> DataFrame:
    """Mutating joins including all rows in x or y.

    See Also:
        [`inner_join()`](datar.dplyr.join.inner_join)
    """
    return _join(
        x,
        y,
        how="outer",
        by=by,
        copy=copy,
        suffix=suffix,
        keep=keep,
    )


@register_verb(
    DataFrame,
    context=Context.EVAL,
    extra_contexts={"by": Context.SELECT},
)
def semi_join(
    x: DataFrame,
    y: DataFrame,
    by: Union[StringOrIter, Mapping[str, str]] = None,
    copy: bool = False,
) -> DataFrame:
    """Returns all rows from x with a match in y.

    See Also:
        [`inner_join()`](datar.dplyr.join.inner_join)
    """
    ret = pandas.merge(
        x,
        y,
        how="left",
        copy=copy,
        suffixes=["", "_y"],
        indicator="__merge__",
        **_merge_on(by),
    )
    ret = ret.loc[ret["__merge__"] == "both", x.columns.tolist()]

    return reconstruct_tibble(x, ret)


@register_verb(
    DataFrame,
    context=Context.EVAL,
    extra_contexts={"by": Context.SELECT},
)
def anti_join(
    x: DataFrame,
    y: DataFrame,
    by: Union[StringOrIter, Mapping[str, str]] = None,
    copy: bool = False,
) -> DataFrame:
    """Returns all rows from x without a match in y.

    See Also:
        [`inner_join()`](datar.dplyr.join.inner_join)
    """
    ret = pandas.merge(
        x,
        y,
        how="left",
        copy=copy,
        suffixes=["", "_y"],
        indicator=True,
        **_merge_on(by),
    )
    ret = ret.loc[ret._merge != "both", x.columns.tolist()]

    return reconstruct_tibble(x, ret)


@register_verb(
    DataFrame,
    context=Context.EVAL,
    extra_contexts={"by": Context.SELECT},
)
def nest_join(
    x: DataFrame,
    y: DataFrame,
    by: Union[StringOrIter, Mapping[str, str]] = None,
    copy: bool = False,
    keep: bool = False,
    name: str = None,
) -> DataFrame:
    """Returns all rows and columns in x with a new nested-df column that
    contains all matches from y

    See Also:
        [`inner_join()`](datar.dplyr.join.inner_join)
    """
    on = by
    if isinstance(by, (list, tuple, set)):
        on = dict(zip(by, by))
    elif by is None:
        common_cols = intersect(
            x.columns.tolist(), y.columns, __calling_env=CallingEnvs.REGULAR
        )
        on = dict(zip(common_cols, common_cols))
    elif not isinstance(by, dict):
        on = {by: by}  # type: ignore

    if copy:
        x = x.copy()

    def get_nested_df(row: Series) -> DataFrame:
        condition = None
        for key in on:
            if condition is None:
                condition = y[on[key]] == row[key]
            else:
                condition = condition & (y[on[key]] == row[key])
        df = filter_(y, condition, __calling_env=CallingEnvs.REGULAR)
        if not keep:
            df = df[
                setdiff(
                    df.columns, on.values(), __calling_env=CallingEnvs.REGULAR
                )
            ]

        return df

    y_matched = x.apply(get_nested_df, axis=1)
    y_name = name or getattr(y, "__dfname__", None)
    if y_name:
        y_matched = y_matched.to_frame(name=y_name)

    out = pandas.concat([x, y_matched], axis=1)

    if isinstance(x, DataFrameGroupBy):
        return x.__class__(
            x,
            _group_vars=group_vars(x, __calling_env=CallingEnvs.REGULAR),
            _group_drop=group_by_drop_default(x),
            _group_data=group_data(x, __calling_env=CallingEnvs.REGULAR),
        )
    return out


def _merge_on(
    by: Union[StringOrIter, Mapping[str, str]]
) -> Mapping[str, StringOrIter]:
    """Calculate argument on for pandas.merge()"""
    if by is None:
        return {}
    if isinstance(by, dict):
        return {"left_on": list(by), "right_on": list(by.values())}
    return {"on": by}



"""Compute lagged or leading values

https://github.com/tidyverse/dplyr/blob/master/R/lead-lag.R
"""
from typing import Iterable, Any

from pandas import Series, Categorical
from pandas.core.dtypes.common import is_categorical_dtype
from pipda import register_func

from ..core.types import NumericType, is_scalar
from ..core.contexts import Context
from ..base import NA
from .order_by import with_order


@register_func(None, context=Context.EVAL)
def lead(
    series: Iterable[Any],
    n: bool = 1,
    default: Any = NA,
    order_by: Iterable[NumericType] = None,
) -> Series:
    """Find next values in a vector

    Args:
        series: Vector of values
        n: Positive integer of length 1, giving the number of positions to
            lead or lag by
        default: Value used for non-existent rows.
        order_by: Override the default ordering to use another vector or column

    Returns:
        Lead or lag values with default values filled to series.
    """
    if n == 0:  # ignore other arguments
        return series

    if order_by is not None:
        return with_order(order_by, lead, series, n=n, default=default)

    series, cats, default = _lead_lag_prepare(series, n, default)
    index = series.index

    ret = default * len(series)
    ret[:-n] = series.values[n:]

    if cats is not None:
        ret = Categorical(ret, categories=cats)
    return Series(ret, index=index)


@register_func(None, context=Context.EVAL)
def lag(
    series: Iterable[Any],
    n: bool = 1,
    default: Any = NA,
    order_by: Iterable[NumericType] = None,
) -> Series:
    """Find previous values in a vector

    See lead()
    """
    if n == 0:  # ignore other arguments
        return series

    if order_by is not None:
        return with_order(order_by, lag, series, n=n, default=default)

    series, cats, default = _lead_lag_prepare(series, n, default)
    index = series.index

    ret = default * len(series)
    ret[n:] = series.values[:-n]

    if cats is not None:
        ret = Categorical(ret, categories=cats)
    return Series(ret, index=index)


def _lead_lag_prepare(data: Iterable[Any], n: int, default: Any):
    """Prepare and check arguments for lead-lag"""
    cats = None
    if is_categorical_dtype(data):
        if isinstance(data, Series):
            cats = data.cat.categories
        else:
            cats = data.categories

    if not isinstance(n, int) or n < 0:
        raise ValueError("`lead-lag` expect a non-negative integer for `n`.")

    if is_scalar(default):
        default = [default]
    if len(default) != 1:
        raise ValueError("`lead-lag` Expect scalar or length-1 `default`.")

    if not isinstance(data, Series):
        data = Series(data)

    return data, cats, default



"""Nest and unnest

https://github.com/tidyverse/tidyr/blob/master/R/nest.R
"""
from typing import Callable, Mapping, Union, Iterable, List
import re

import pandas
from pandas import DataFrame, Series
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.types import Dtype, is_scalar
from ..core.utils import vars_select, recycle_value, to_df, reconstruct_tibble
from ..core.grouped import DataFrameGroupBy, DataFrameRowwise
from ..core.contexts import Context

from ..base import setdiff, NA
from ..dplyr import distinct, bind_cols, group_vars, group_by_drop_default

from .chop import unchop, _vec_split
from .pack import unpack


@register_verb(DataFrame, context=Context.SELECT)
def nest(
    _data: DataFrame,
    _names_sep: str = None,
    base0_: bool = None,
    **cols: Union[str, int],
) -> DataFrame:
    """Nesting creates a list-column of data frames

    Args:
        _data: A data frame
        **cols: Columns to nest
        _names_sep: If `None`, the default, the names will be left as is.
            Inner names will come from the former outer names
            If a string, the inner and outer names will be used together.
            The names of the new outer columns will be formed by pasting
            together the outer and the inner column names, separated by
            `_names_sep`.
        base0_: Whether `**cols` are 0-based
            if not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        Nested data frame.
    """
    if not cols:
        raise ValueError("`**cols` must not be empty.")

    all_columns = _data.columns
    colgroups = {}
    usedcols = set()
    for group, columns in cols.items():
        old_cols = all_columns[vars_select(all_columns, columns, base0=base0_)]
        usedcols = usedcols.union(old_cols)
        newcols = (
            old_cols
            if _names_sep is None
            else _strip_names(old_cols, group, _names_sep)
        )
        colgroups[group] = dict(zip(newcols, old_cols))

    asis = setdiff(_data.columns, usedcols, __calling_env=CallingEnvs.REGULAR)
    keys = _data[asis]
    u_keys = distinct(keys, __calling_env=CallingEnvs.REGULAR)
    nested = []
    for group, columns in colgroups.items():
        if _names_sep is None:  # names as is
            # out <- map(cols, ~ vec_split(.data[.x], keys)$val)
            val = _vec_split(_data[list(columns)], keys).val
        else:
            # out <- map(
            #   cols,
            #   ~ vec_split(set_names(.data[.x], names(.x)), keys)$val
            # )
            to_split = _data[list(columns.values())]
            to_split.columns = list(columns)
            val = _vec_split(to_split, keys).val
        nested.append(val.reset_index(drop=True))

    out = pandas.concat(nested, ignore_index=True, axis=1)
    out.columns = list(colgroups)
    if u_keys.shape[1] == 0:
        return out if isinstance(out, DataFrame) else out.to_frame()
    return bind_cols(
        u_keys,
        recycle_value(out, u_keys.shape[0]),
        __calling_env=CallingEnvs.REGULAR,
    )


@nest.register(DataFrameGroupBy, context=Context.SELECT)
def _(
    _data: DataFrameGroupBy,
    _names_sep: str = None,
    base0_: bool = None,
    **cols: Mapping[str, Union[str, int]],
) -> DataFrameGroupBy:
    """Nesting grouped dataframe"""
    if not cols:
        cols = {
            "data": setdiff(
                _data.columns,
                group_vars(_data, __calling_env=CallingEnvs.REGULAR),
                __calling_env=CallingEnvs.REGULAR,
            )
        }
    out = nest.dispatch(DataFrame)(
        _data, **cols, _names_sep=_names_sep, base0_=base0_
    )
    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_verb(DataFrame, context=Context.SELECT)
def unnest(
    data: DataFrame,
    *cols: Union[str, int],
    keep_empty: bool = False,
    ptype: Union[Dtype, Mapping[str, Dtype]] = None,
    names_sep: str = None,
    names_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
) -> DataFrame:
    """Flattens list-column of data frames back out into regular columns.

    Args:
        data: A data frame to flatten.
        *cols: Columns to unnest.
        keep_empty: By default, you get one row of output for each element
            of the list your unchopping/unnesting.
            This means that if there's a size-0 element
            (like NULL or an empty data frame), that entire row will be
            dropped from the output.
            If you want to preserve all rows, use `keep_empty` = `True` to
            replace size-0 elements with a single row of missing values.
        ptype: Providing the dtypes for the output columns.
            Could be a single dtype, which will be applied to all columns, or
            a dictionary of dtypes with keys for the columns and values the
            dtypes.
        names_sep: If `None`, the default, the names will be left as is.
            Inner names will come from the former outer names
            If a string, the inner and outer names will be used together.
            The names of the new outer columns will be formed by pasting
            together the outer and the inner column names, separated by
            `names_sep`.
        names_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether `cols` are 0-based
            if not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        Data frame with selected columns unnested.
    """
    if not cols:
        raise ValueError("`*cols` is required when using unnest().")

    all_columns = data.columns
    cols = vars_select(all_columns, cols, base0=base0_)
    cols = all_columns[cols]

    if isinstance(data, DataFrameGroupBy):
        out = data.copy(copy_grouped=True)
    else:
        out = data.copy()

    for col in cols:
        out[col] = _as_df(data[col])

    out = unchop(
        out,
        cols,
        keep_empty=keep_empty,
        ptype=ptype,
        base0_=base0_,
        __calling_env=CallingEnvs.REGULAR,
    )
    return unpack(
        out,
        cols,
        names_sep=names_sep,
        names_repair=names_repair,
        __calling_env=CallingEnvs.REGULAR,
    )


@unnest.register(DataFrameRowwise, context=Context.SELECT)
def _(
    data: DataFrameRowwise,
    *cols: Union[str, int],
    keep_empty: bool = False,
    ptype: Union[Dtype, Mapping[str, Dtype]] = None,
    names_sep: str = None,
    names_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
) -> DataFrameGroupBy:
    """Unnest rowwise dataframe"""
    out = unnest.dispatch(DataFrame)(
        data,
        *cols,
        keep_empty=keep_empty,
        ptype=ptype,
        names_sep=names_sep,
        names_repair=names_repair,
        base0_=base0_,
    )
    return DataFrameGroupBy(
        out,
        _group_vars=group_vars(data, __calling_env=CallingEnvs.REGULAR),
        _group_drop=group_by_drop_default(data),
    )


def _strip_names(names: Iterable[str], base: str, sep: str) -> List[str]:
    """Strip the base names with sep"""
    out = []
    for name in names:
        if not sep:
            out.append(name[len(base) :] if name.startswith(base) else name)
        else:
            parts = re.split(re.escape(sep), name, maxsplit=1)
            out.append(parts[1] if parts[0] == base else name)
    return out


def _as_df(series: Series) -> List[DataFrame]:
    """Convert series to dataframe"""
    out = []
    for val in series:
        if isinstance(val, DataFrame):
            if val.shape[1] == 0:  # no columns
                out.append(NA)
            elif val.shape[0] == 0:
                out.append(
                    DataFrame([[NA] * val.shape[1]], columns=val.columns)
                )
            else:
                out.append(val)
        elif is_scalar(val) and pandas.isnull(val):
            out.append(val)
        else:
            out.append(to_df(val, name=series.name))
    return out



"""Count observations by group

See souce code https://github.com/tidyverse/dplyr/blob/master/R/count-tally.R
"""
from typing import Any, Iterable, Union

from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs
from pipda.expression import Expression

from ..core.contexts import Context
from ..core.types import NumericOrIter, NumericType
from ..core.utils import copy_attrs, logger
from ..core.defaults import f
from ..core.grouped import DataFrameGroupBy
from ..base import options_context, sum_
from .context import n
from .group_by import group_by_drop_default, group_by
from .group_data import group_data, group_vars
from .mutate import mutate
from .summarise import summarise
from .arrange import arrange
from .desc import desc


@register_verb(DataFrame, context=Context.PENDING)
def count(
    x: DataFrame,
    *args: Any,
    wt: NumericOrIter = None,
    sort: bool = False,
    name: str = None,
    _drop: bool = None,
    **kwargs: Any,
) -> DataFrame:
    """Count observations by group

    See https://dplyr.tidyverse.org/reference/count.html

    Args:
        x: The dataframe
        *columns: and
        **mutates: Variables to group by
        wt: Frequency weights. Can be None or a variable:
            If None (the default), counts the number of rows in each group.
            If a variable, computes sum(wt) for each group.
        sort: If TRUE, will show the largest groups at the top.
        name: The name of the new column in the output.

    Returns:
        DataFrame object with the count column
    """
    if _drop is None:
        _drop = group_by_drop_default(x)

    if args or kwargs:
        out = group_by(
            x,
            *args,
            **kwargs,
            _add=True,
            _drop=_drop,
            __calling_env=CallingEnvs.REGULAR,
        )
    else:
        out = x

    out = tally(
        out,
        wt=wt,
        sort=sort,
        name=name,
        __calling_env=CallingEnvs.REGULAR,
    )
    if isinstance(x, DataFrameGroupBy):
        out = DataFrameGroupBy(
            out,
            _group_vars=group_vars(x, __calling_env=CallingEnvs.REGULAR),
            _group_drop=_drop,
            _group_data=group_data(x, __calling_env=CallingEnvs.REGULAR),
        )
    return out


@register_verb(DataFrame, context=Context.PENDING)
def tally(
    x: DataFrame,
    wt: NumericOrIter = None,
    sort: bool = False,
    name: str = None,
) -> DataFrame:
    """A ower-level function for count that assumes you've done the grouping

    See count()
    """
    tallyn = _tally_n(wt)

    name = _check_name(name, group_vars(x, __calling_env=CallingEnvs.REGULAR))
    # thread-safety?
    with options_context(dplyr_summarise_inform=False):

        out = summarise(
            x,
            {name: n() if tallyn is None else tallyn},
            __calling_env=CallingEnvs.REGULAR,
        )

    # keep attributes
    copy_attrs(out, x)

    if sort:
        return arrange(out, desc(f[name]), __calling_env=CallingEnvs.REGULAR)
    return out


@register_verb(DataFrame, context=Context.PENDING)
def add_count(
    x: DataFrame,
    *args: Any,
    wt: str = None,
    sort: bool = False,
    name: str = "n",
    **kwargs: Any,
) -> DataFrame:
    """Equivalents to count() but use mutate() instead of summarise()

    See count().
    """
    if args or kwargs:
        out = group_by(
            x,
            *args,
            **kwargs,
            _add=True,
            __calling_env=CallingEnvs.REGULAR,
        )
    else:
        out = x

    out = add_tally(
        out, wt=wt, sort=sort, name=name, __calling_env=CallingEnvs.REGULAR
    )
    return out


@register_verb(DataFrame, context=Context.PENDING)
def add_tally(
    x: DataFrame, wt: str = None, sort: bool = False, name: str = "n"
) -> DataFrame:
    """Equivalents to tally() but use mutate() instead of summarise()

    See count().
    """
    tallyn = _tally_n(wt)
    name = _check_name(name, x.columns)

    out = mutate(
        x,
        {name: n() if tallyn is None else tallyn},
        __calling_env=CallingEnvs.REGULAR,
    )

    if sort:
        return arrange(
            out,
            desc(f[name]),
            __calling_env=CallingEnvs.REGULAR,
        )
    return out


# Helpers -----------------------------------------------------------------
def _tally_n(wt: Union[NumericOrIter, Expression]) -> Iterable[NumericType]:
    """Compuate the weights for counting"""
    if wt is None:
        return None  # will be n() later on

    # If it's Expression, will return a Function object
    # Otherwise, sum of wt

    return sum_(wt, na_rm=True, __calling_env=CallingEnvs.PIPING)


def _check_name(name: str, invars: Iterable[str]) -> str:
    """Check if count is valid"""
    if name is None:
        name = _n_name(invars)

        if name != "n":
            logger.warning(
                "Storing counts in `%s`, as `n` already present in input. "
                'Use `name="new_name" to pick a new name.`'
            )
    elif not isinstance(name, str):
        raise ValueError("`name` must be a single string.")

    return name


def _n_name(invars: Iterable[str]) -> str:
    """Make sure that name does not exist in invars"""
    name = "n"
    while name in invars:
        name = "n" + name
    return name



"""Constants and functions about NULL (none)"""

from typing import Any

from pipda import register_func

from ..core.contexts import Context


NULL = None


@register_func(None, context=Context.EVAL)
def as_null(*args: Any, **kwargs: Any) -> None:
    """Ignores arguments and returns NULL (None). R's `as.null()`

    Args:
        *args: and
        **kwargs: arguments to be ignored

    Returns:
        `NULL`
    """
    return NULL


@register_func(None, context=Context.EVAL)
def is_null(x: Any) -> bool:
    """Check if x is exactly NULL (None), same as R's `is.null()`

    Args:
        x: The value to check

    Returns:
        True if x is NULL (None) else False
    """
    return x is NULL



"""Iterate over groups

https://github.com/tidyverse/dplyr/blob/master/R/group_split.R
https://github.com/tidyverse/dplyr/blob/master/R/group_map.R
https://github.com/tidyverse/dplyr/blob/master/R/group_trim.R
"""

from typing import Any, Callable, Iterable, List

import pandas
from pandas import DataFrame
from pipda import register_verb, evaluate_expr
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.grouped import DataFrameGroupBy, DataFrameRowwise
from ..core.utils import (
    copy_attrs,
    logger,
    vars_select,
    nargs,
    reconstruct_tibble,
)
from ..core.types import StringOrIter
from ..base import is_factor, droplevels, setdiff, intersect
from .group_data import group_keys, group_rows, group_vars
from .group_by import group_by, ungroup
from .select import select
from .mutate import mutate
from .across import across
from .tidyselect import where


@register_verb(DataFrame, context=Context.EVAL)
def group_map(
    _data: DataFrame,
    _f: Callable,
    *args: Any,
    _keep: bool = False,
    **kwargs: Any,
) -> Iterable:
    """A generator to map function to data in each group"""
    keys = (
        group_keys(_data, __calling_env=CallingEnvs.REGULAR)
        if nargs(_f) > 1
        else None
    )
    for i, chunk in enumerate(
        group_split(_data, _keep=_keep, __calling_env=CallingEnvs.REGULAR)
    ):
        if keys is None:
            yield _f(chunk)
        else:
            yield _f(chunk, keys.iloc[[i], :], *args, **kwargs)


def _group_map_list(
    _data: DataFrame,
    _f: Callable,
    *args: Any,
    _keep: bool = False,
    **kwargs: Any,
) -> List:
    """List version of group_map"""
    return list(
        group_map(
            _data,
            _f,
            *args,
            _keep=_keep,
            **kwargs,
            __calling_env=CallingEnvs.REGULAR,
        )
    )


group_map.list = register_verb(DataFrame, context=Context.PENDING)(
    _group_map_list
)


@register_verb(DataFrame, context=Context.EVAL)
def group_modify(
    _data: DataFrame,
    _f: Callable,
    *args: Any,
    _keep: bool = False,
    **kwargs: Any,
) -> DataFrame:
    """Modify data in each group with func, returns a dataframe"""
    return _f(_data, *args, **kwargs)


@group_modify.register(DataFrameGroupBy)
def _(
    _data: DataFrameGroupBy,
    _f: Callable,
    *args: Any,
    _keep: bool = False,
    **kwargs: Any,
) -> DataFrameGroupBy:
    gvars = group_vars(_data, __calling_env=CallingEnvs.REGULAR)
    func = (lambda df, keys: _f(df)) if nargs(_f) == 1 else _f

    def fun(df, keys):
        res = func(df, keys, *args, **kwargs)
        if not isinstance(res, DataFrame):
            raise ValueError("The result of `_f` should be a data frame.")
        bad = intersect(res.columns, gvars, __calling_env=CallingEnvs.REGULAR)
        if bad:
            raise ValueError(
                "The returned data frame cannot contain the original grouping "
                f"variables: {bad}."
            )

        return pandas.concat(
            (
                keys.iloc[[0] * res.shape[0], :].reset_index(drop=True),
                res.reset_index(drop=True),
            ),
            axis=1,
        )

    chunks = group_map(
        _data, fun, _keep=_keep, __calling_env=CallingEnvs.REGULAR
    )
    out = pandas.concat(chunks, axis=0)

    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_verb(DataFrame, context=Context.EVAL)
def group_walk(
    _data: DataFrame,
    _f: Callable,
    *args: Any,
    _keep: bool = False,
    **kwargs: Any,
) -> None:
    """Walk along data in each groups, but don't return anything"""
    # list to trigger generator
    list(
        group_map(
            _data,
            _f,
            _keep=_keep,
            *args,
            **kwargs,
            __calling_env=CallingEnvs.REGULAR,
        )
    )


@register_verb(DataFrame)
def group_trim(_data: DataFrame, _drop: bool = None) -> DataFrame:
    """Trim the unused group levels"""
    return _data


@group_trim.register(DataFrameGroupBy)
def _(_data: DataFrame, _drop: bool = None) -> DataFrameGroupBy:
    """Group trim on grouped data"""
    ungrouped = ungroup(_data, __calling_env=CallingEnvs.REGULAR)

    fgroups = select(
        ungrouped, where(is_factor), __calling_env=CallingEnvs.REGULAR
    )
    dropped = mutate(
        ungrouped,
        across(fgroups.columns.tolist(), droplevels),
        __calling_env=CallingEnvs.REGULAR,
    )

    return reconstruct_tibble(_data, dropped, keep_rowwise=True)


@register_verb(DataFrame, context=Context.PENDING)
def with_groups(
    _data: DataFrame,
    _groups: StringOrIter,
    _func: Callable,
    *args: Any,
    **kwargs: Any,
) -> DataFrameGroupBy:
    """Modify the grouping variables for a single operation.

    Args:
        _data: A data frame
        _groups: columns passed by group_by
            Use None to temporarily ungroup.
        _func: Function to apply to regrouped data.

    Returns:
        The new data frame with operations applied.
    """
    if _groups is None:
        grouped = ungroup(_data, __calling_env=CallingEnvs.REGULAR)
    else:
        all_columns = _data.columns
        _groups = evaluate_expr(_groups, _data, Context.SELECT)
        _groups = all_columns[vars_select(all_columns, _groups)]
        grouped = group_by(_data, *_groups, __calling_env=CallingEnvs.REGULAR)

    out = _func(grouped, *args, **kwargs)
    copy_attrs(out, _data)
    return out


@register_verb(DataFrame, context=Context.EVAL)
def group_split(
    _data: DataFrame, *args: Any, _keep: bool = True, **kwargs: Any
) -> Iterable[DataFrame]:
    """Get a list of data in each group"""
    data = group_by(_data, *args, **kwargs, __calling_env=CallingEnvs.REGULAR)
    yield from group_split_impl(data, _keep=_keep)


@group_split.register(DataFrameGroupBy, context=Context.EVAL)
def _(
    _data: DataFrame, *args: Any, _keep: bool = True, **kwargs: Any
) -> DataFrameGroupBy:
    # data = group_by(_data, *args, **kwargs, _add=True)
    if args or kwargs:
        logger.warning(
            "`*args` and `**kwargs` are ignored in "
            "`group_split(<DataFrameGroupBy)`, please use "
            "`group_by(..., _add=True) >> group_split()`."
        )
    return group_split_impl(_data, _keep=_keep)


@group_split.register(DataFrameRowwise, context=Context.EVAL)
def _(
    _data: DataFrame, *args: Any, _keep: bool = None, **kwargs: Any
) -> DataFrameRowwise:
    if args or kwargs:
        logger.warning(
            "`*args` and `**kwargs` is ignored in "
            "`group_split(<DataFrameRowwise>)`."
        )
    if _keep is not None:
        logger.warning(
            "`_keep` is ignored in " "`group_split(<DataFrameRowwise>)`."
        )

    return group_split_impl(_data, _keep=True)


def _group_split_list(
    _data: DataFrame, *args: Any, _keep: bool = True, **kwargs: Any
) -> Iterable[DataFrame]:
    """List version of group_split"""
    return list(
        group_split(
            _data,
            *args,
            _keep=_keep,
            **kwargs,
            __calling_env=CallingEnvs.REGULAR,
        )
    )


group_split.list = register_verb(DataFrame, context=Context.PENDING)(
    _group_split_list
)


def group_split_impl(data: DataFrame, _keep: bool):
    """Implement splitting data frame by groups"""
    out = ungroup(data, __calling_env=CallingEnvs.REGULAR)
    indices = group_rows(data, __calling_env=CallingEnvs.REGULAR)

    if not _keep:
        remove = group_vars(data, __calling_env=CallingEnvs.REGULAR)
        _keep = out.columns
        _keep = setdiff(_keep, remove, __calling_env=CallingEnvs.REGULAR)
        out = out[_keep]

    for rows in indices:
        yield out.iloc[rows, :]



"""Testing functions in R"""

# all_equal, all_equal_numeric, ...
# all, any
# is_true, is_false
# is_finite, is_infinite, is_nan
# is_function, is_atomic, is_vector, is_array, is_character, is_complex,
# is_data_frame, is_double, is_element, (is_factor, is_ordered in base.factor)
# is_integer, is_list, is_logical, is_numeric
# is_unsorted, (is_qr in base.qr)
import builtins
from typing import Any, Callable, Iterable

import numpy
from pandas.core.dtypes.common import (
    is_integer_dtype,
    is_float_dtype,
    is_numeric_dtype,
)
from pipda import register_func

from ..core.contexts import Context
from ..core.types import TypeOrIter, BoolOrIter, is_scalar


def _register_type_testing(
    name: str,
    scalar_types: TypeOrIter,
    dtype_checker: Callable[[Iterable], bool],
    doc: str = "",
) -> Callable:
    """Register type testing function"""

    @register_func(None, context=Context.EVAL)
    def _testing(x: Any) -> bool:
        """Type testing"""
        if is_scalar(x):
            return isinstance(x, scalar_types)
        if hasattr(x, "dtype"):
            return dtype_checker(x)
        return all(isinstance(elem, scalar_types) for elem in x)

    _testing.__name__ = name
    _testing.__doc__ = doc
    return _testing


is_numeric = _register_type_testing(
    "is_numeric",
    scalar_types=(int, float, complex, numpy.number),
    dtype_checker=is_numeric_dtype,
    doc="""Test if a value is numeric

    Args:
        x: The value to be checked

    Returns:
        True if the value is numeric; with a numeric dtype;
        or all elements are numeric
    """,
)

is_integer = _register_type_testing(
    "is_integer",
    scalar_types=(int, numpy.integer),
    dtype_checker=is_integer_dtype,
    doc="""Test if a value is integers

    Alias `is_int`

    Args:
        x: The value to be checked

    Returns:
        True if the value is an integer or integers; False otherwise.
    """,
)

is_int = is_integer

is_double = _register_type_testing(
    "is_double",
    scalar_types=(float, numpy.float_),
    dtype_checker=is_float_dtype,
    doc="""Test if a value is integers

    Alias `is_float`

    Args:
        x: The value to be checked

    Returns:
        True if the value is an integer or integers; False otherwise.
    """,
)

is_float = is_double


@register_func(None, context=Context.EVAL)
def is_atomic(x: Any) -> bool:
    """Check if x is an atomic or scalar value

    Args:
        x: The value to be checked

    Returns:
        True if x is atomic otherwise False
    """
    return is_scalar(x)


@register_func(None, context=Context.EVAL)
def is_element(elem: Any, elems: Iterable[Any]) -> BoolOrIter:
    """R's `is.element()` or `%in%`.

    Alias `is_in()`

    We can't do `a %in% b` in python (`in` behaves differently), so
    use this function instead
    """
    out = numpy.in1d(elem, elems)
    if is_scalar(elem):  # necessary?
        return bool(out)
    return out


is_in = is_element


all = register_func(None, context=Context.EVAL)(
    # can't set attributes to builtins.all, so wrap it.
    lambda arg: builtins.all(arg)
)
any = register_func(None, context=Context.EVAL)(
    lambda arg: builtins.any(arg)
)



"""Provides other helper functions for factors"""
from typing import Any, Iterable
import numpy
from pandas import Categorical, DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.types import ForcatsRegType, ForcatsType, is_null, is_scalar
from ..core.utils import Array
from ..core.contexts import Context

from ..core.defaults import f
from ..base import (
    factor,
    tabulate,
    prop_table,
    nlevels,
    levels,
    NA,
    setdiff,
    is_ordered,
)
from ..dplyr import arrange, desc, mutate

from .utils import check_factor
from .lvl_order import fct_inorder


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_count(_f: ForcatsType, sort: bool = False, prop=False) -> Categorical:
    """Count entries in a factor

    Args:
        _f: A factor
        sort: If True, sort the result so that the most common values float to
            the top
        prop: If True, compute the fraction of marginal table.

    Returns:
        A data frame with columns `f`, `n` and `p`, if prop is True
    """
    f2 = check_factor(_f)
    n_na = sum(is_null(f2))

    df = DataFrame(
        {
            "f": fct_inorder(
                levels(f2, __calling_env=CallingEnvs.REGULAR),
                __calling_env=CallingEnvs.REGULAR,
            ),
            "n": tabulate(
                f2,
                nlevels(f2, __calling_env=CallingEnvs.REGULAR),
                __calling_env=CallingEnvs.REGULAR,
            ),
        }
    )

    if n_na > 0:
        df = df.append({"f": NA, "n": n_na}, ignore_index=True)

    if sort:
        df = arrange(
            df,
            desc(f.n, __calling_env=CallingEnvs.PIPING),
            __calling_env=CallingEnvs.REGULAR,
        )
    if prop:
        df = mutate(
            df,
            p=prop_table(f.n, __calling_env=CallingEnvs.PIPING),
            __calling_env=CallingEnvs.REGULAR,
        )

    return df


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_match(_f: ForcatsType, lvls: Any) -> Iterable[bool]:
    """Test for presence of levels in a factor

    Do any of `lvls` occur in `_f`?

    Args:
        _f: A factor
        lvls: A vector specifying levels to look for.

    Returns:
        A logical factor
    """
    _f = check_factor(_f)

    if is_scalar(lvls):
        lvls = [lvls]

    bad_lvls = setdiff(
        lvls,
        levels(_f, __calling_env=CallingEnvs.REGULAR),
        __calling_env=CallingEnvs.REGULAR,
    )
    if len(bad_lvls) > 0:
        bad_lvls = Array(bad_lvls)[~is_null(bad_lvls)]
    if len(bad_lvls) > 0:
        raise ValueError(f"Levels not present in factor: {bad_lvls}.")

    return numpy.isin(_f, lvls)


@register_verb(ForcatsRegType)
def fct_unique(_f: ForcatsType) -> Categorical:
    """Unique values of a factor

    Args:
        _f: A factor

    Returns:
        The factor with the unique values in `_f`
    """
    lvls = levels(_f, __calling_env=CallingEnvs.REGULAR)
    is_ord = is_ordered(_f, __calling_env=CallingEnvs.REGULAR)
    return factor(lvls, lvls, exclude=None, ordered=is_ord)



"""Unite multiple columns into one by pasting strings together"""

from typing import Union

import pandas
from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.utils import vars_select, reconstruct_tibble

from ..base import setdiff


@register_verb(DataFrame, context=Context.SELECT)
def unite(
    data: DataFrame,
    col: str,
    *columns: Union[str, int],
    sep: str = "_",
    remove: bool = True,
    na_rm: bool = False,
    base0_: bool = None,
) -> DataFrame:
    """Unite multiple columns into one by pasting strings together

    Args:
        data: A data frame.
        col: The name of the new column, as a string or symbol.
        *columns: Columns to unite
        sep: Separator to use between values.
        remove: If True, remove input columns from output data frame.
        na_rm: If True, missing values will be remove prior to uniting
            each value.
        base0_: Whether `columns` is 0-based when given by index
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The dataframe with selected columns united
    """
    all_columns = data.columns
    if not columns:
        columns = all_columns
    else:
        columns = all_columns[vars_select(all_columns, *columns, base0=base0_)]

    out = data.copy()

    def unite_cols(row):
        if na_rm:
            row = [elem for elem in row if not pandas.isnull(elem)]
        return sep.join(str(elem) for elem in row)

    out[col] = out[columns].agg(unite_cols, axis=1)
    # get indexes to relocate
    insert_at = min(data.columns.get_indexer_for(columns))
    relocated_cols = (
        data.columns[:insert_at]
        .difference([col])
        .union([col], sort=False)
        .union(data.columns[insert_at:].difference([col]), sort=False)
    )
    out = out[relocated_cols]

    if remove:
        cols_to_remove = setdiff(
            columns,
            [col],
            __calling_env=CallingEnvs.REGULAR
        )
        if len(cols_to_remove) > 0:
            out.drop(columns=cols_to_remove, inplace=True)

    return reconstruct_tibble(data, out)



"""Extract a character column into multiple columns using regular
expression groups

https://github.com/tidyverse/tidyr/blob/HEAD/R/extract.R
"""
import re
from typing import Union, Mapping

import pandas
from pandas import DataFrame
from pipda import register_verb

from ..core.types import StringOrIter, Dtype, is_scalar
from ..core.contexts import Context
from ..core.utils import apply_dtypes, vars_select, reconstruct_tibble


@register_verb(DataFrame, context=Context.SELECT)
def extract(
    data: DataFrame,
    col: Union[str, int],
    into: StringOrIter,
    regex: str = r"(\w+)",
    remove: bool = True,
    convert: Union[bool, Dtype, Mapping[str, Dtype]] = False,
    base0_: bool = None,
) -> DataFrame:
    """Given a regular expression with capturing groups, extract() turns each
    group into a new column. If the groups don't match, or the input is NA,
    the output will be NA.

    See https://tidyr.tidyverse.org/reference/extract.html

    Args:
        data: The dataframe
        col: Column name or position.
        into: Names of new variables to create as character vector.
            Use None to omit the variable in the output.
        regex: a regular expression used to extract the desired values.
            There should be one group (defined by ()) for each element of into.
        remove: If TRUE, remove input column from output data frame.
        convert: The universal type for the extracted columns or a dict for
            individual ones
        base0_: Whether `col` is 0-based when given by index
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        Dataframe with extracted columns.
    """
    if is_scalar(into):
        into = [into]  # type: ignore

    all_columns = data.columns
    col = vars_select(all_columns, col, base0=base0_)
    col = all_columns[col[0]]

    outcols = {}
    # merge columns with same name
    # all columns are already strs
    # # 'col' => i, j, k
    # # i, j, k are indexes that have same name 'col'
    mergedcols = {}
    for i, outcol in enumerate(into):
        if is_scalar(outcol) and pandas.isnull(outcol):
            continue
        if not isinstance(outcol, str):
            raise ValueError(
                "`into` must be a string or an iterable of strings."
            )
        outcols[i] = outcol
        mergedcols.setdefault(outcol, []).append(i)

    regex = re.compile(regex)
    if regex.groups != len(into):
        raise ValueError(
            f"`regex` should define {len(into)} groups; "
            f"found {regex.groups}."
        )
    out = data[col].str.extract(regex)
    out = {
        outcol: (
            out.iloc[:, indexes[0]]
            if len(indexes) == 1
            else out.iloc[:, indexes].astype(str).agg("".join, axis=1)
        )
        for outcol, indexes in mergedcols.items()
    }

    out = DataFrame(out)
    apply_dtypes(out, convert)

    base = data[all_columns.difference([col])] if remove else data
    out = pandas.concat([base, out], axis=1)
    return reconstruct_tibble(data, out, keep_rowwise=True)



"""Set operations

https://github.com/tidyverse/dplyr/blob/master/R/sets.r
"""

import pandas
from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.grouped import DataFrameGroupBy
from ..core.utils import reconstruct_tibble

from ..base.verbs import intersect, union, setdiff, setequal
from .bind import bind_rows


def _check_xy(x: DataFrame, y: DataFrame) -> None:
    """Check the dimension and columns of x and y for set operations"""
    if x.shape[1] != y.shape[1]:
        raise ValueError(
            "not compatible:\n"
            f"- different number of columns: {x.shape[1]} vs {y.shape[1]}"
        )

    in_y_not_x = setdiff(
        y.columns, x.columns, __calling_env=CallingEnvs.REGULAR
    )
    in_x_not_y = setdiff(
        x.columns, y.columns, __calling_env=CallingEnvs.REGULAR
    )
    if in_y_not_x or in_x_not_y:
        msg = ["not compatible:"]
        if in_y_not_x:
            msg.append(f"- Cols in `y` but not `x`: {in_y_not_x}.")
        if in_x_not_y:
            msg.append(f"- Cols in `x` but not `y`: {in_x_not_y}.")
        raise ValueError("\n".join(msg))


@intersect.register(DataFrame, context=Context.EVAL)
def _(x: DataFrame, y: DataFrame) -> DataFrame:
    """Intersect of two dataframes

    Args:
        _data, data2, *datas: Dataframes to perform operations
        on: The columns to the dataframes to perform operations on

    Returns:
        The dataframe of intersect of input dataframes
    """
    _check_xy(x, y)
    from .distinct import distinct

    return distinct(
        pandas.merge(x, y, how="inner"), __calling_env=CallingEnvs.REGULAR
    )


@intersect.register(DataFrameGroupBy, context=Context.EVAL)
def _(x: DataFrameGroupBy, y: DataFrame) -> DataFrameGroupBy:
    out = intersect.dispatch(DataFrame)(x, y)
    return reconstruct_tibble(x, out, keep_rowwise=True)


@union.register(DataFrame, context=Context.EVAL)
def _(x: DataFrame, y: DataFrame) -> DataFrame:
    """Union of two dataframes

    Args:
        _data, data2, *datas: Dataframes to perform operations
        on: The columns to the dataframes to perform operations on

    Returns:
        The dataframe of union of input dataframes
    """
    _check_xy(x, y)
    from .distinct import distinct

    return distinct(
        pandas.merge(x, y, how="outer"), __calling_env=CallingEnvs.REGULAR
    )


@union.register(DataFrameGroupBy, context=Context.EVAL)
def _(x: DataFrameGroupBy, y: DataFrame) -> DataFrameGroupBy:
    out = union.dispatch(DataFrame)(x, y)
    return reconstruct_tibble(x, out, keep_rowwise=True)


@setdiff.register(DataFrame, context=Context.EVAL)
def _(x: DataFrame, y: DataFrame) -> DataFrame:
    """Set diff of two dataframes

    Args:
        _data, *datas: Dataframes to perform operations
        on: The columns to the dataframes to perform operations on

    Returns:
        The dataframe of setdiff of input dataframes
    """
    _check_xy(x, y)
    indicator = "__datar_setdiff__"
    out = pandas.merge(x, y, how="left", indicator=indicator)

    from .distinct import distinct

    return distinct(
        out[out[indicator] == "left_only"]
        .drop(columns=[indicator])
        .reset_index(drop=True),
        __calling_env=CallingEnvs.REGULAR,
    )


@setdiff.register(DataFrameGroupBy, context=Context.EVAL)
def _(x: DataFrameGroupBy, y: DataFrame) -> DataFrameGroupBy:
    out = setdiff.dispatch(DataFrame)(x, y)
    return reconstruct_tibble(x, out, keep_rowwise=True)


@register_verb(DataFrame, context=Context.EVAL)
def union_all(x: DataFrame, y: DataFrame) -> DataFrame:
    """Union of all rows of two dataframes

    Args:
        _data, *datas: Dataframes to perform operations
        on: The columns to the dataframes to perform operations on

    Returns:
        The dataframe of union of all rows of input dataframes
    """
    _check_xy(x, y)
    return bind_rows(x, y, __calling_env=CallingEnvs.REGULAR)


@union_all.register(DataFrameGroupBy, context=Context.EVAL)
def _(x: DataFrameGroupBy, y: DataFrame) -> DataFrameGroupBy:
    out = union_all.dispatch(DataFrame)(x, y)
    return reconstruct_tibble(x, out, keep_rowwise=True)


@setequal.register(DataFrame, context=Context.EVAL)
def _(x: DataFrame, y: DataFrame) -> bool:
    """Check if two dataframes equal

    Args:
        _data: The first dataframe
        data2: The second dataframe

    Returns:
        True if they equal else False
    """
    _check_xy(x, y)

    x = x.sort_values(by=x.columns.to_list()).reset_index(drop=True)
    y = y.sort_values(by=y.columns.to_list()).reset_index(drop=True)
    return x.equals(y)



"""Cast values between types"""

from typing import Any

import numpy
from pipda import register_func

from ..core.types import (
    Dtype,
    NumericOrIter,
    DoubleOrIter,
    IntOrIter,
    is_scalar,
    is_categorical,
)
from ..core.contexts import Context
from ..core.utils import categorized, get_option
from ..core.types import is_null

from .na import NA


def _as_type(x: Any, type_: Dtype, na: Any = None) -> Any:
    """Convert x or elements of x to certain type"""
    if is_scalar(x):
        if is_null(x) and na is not None:
            return na
        return type_(x)  # type: ignore

    if hasattr(x, "astype"):
        if na is None:
            return x.astype(type_)

        na_mask = is_null(x)
        out = x.astype(type_)

        try:
            out[na_mask] = na
        except (ValueError, TypeError):
            out = out.astype(object)
            out[na_mask] = na

        return out

    return type(x)([_as_type(elem, type_=type_, na=na) for elem in x])


@register_func(None, context=Context.EVAL)
def as_double(x: Any) -> DoubleOrIter:
    """Convert an object or elements of an iterable into double/float

    Args:
        x: The object

    Returns:
        Values converted into numpy.float64
    """
    return _as_type(x, numpy.double)


@register_func(None, context=Context.EVAL)
def as_float(x: Any, float_dtype: Dtype = numpy.float_) -> DoubleOrIter:
    """Convert an object or elements of an iterable into double/float

    Args:
        x: The object

    Returns:
        Converted value according to the float dtype
    """
    return _as_type(x, float_dtype)


@register_func(None, context=Context.EVAL)
def as_integer(
    x: Any,
    integer_dtype: Dtype = numpy.int_,
    _keep_na: bool = True,
    base0_: bool = None,
) -> IntOrIter:
    """Convert an object or elements of an iterable into int64

    Alias `as_int`

    Args:
        x: The object
        integer_dtype: The dtype of the integer. Could be one of:
            - `numpy.int0`
            - `numpy.int16`
            - `numpy.int32`
            - `numpy.int64`
            - `numpy.int8`
            - `numpy.int_`
            - `numpy.intc`
            - `numpy.intp`
        _keep_na: If True, NAs will be kept, then the dtype will be object
            (interger_dtype ignored)
        base0_: When converting factors with strings, whether convert them into
            0-based codes or not.
            If not provided, will use `get_option('which.base.0')`

    Returns:
        Converted values according to the integer_dtype
    """
    if is_categorical(x):
        base0_ = get_option("which.base.0", base0_)
        return categorized(x).codes + int(not base0_)
    return _as_type(x, integer_dtype, na=NA if _keep_na else None)


as_int = as_integer


@register_func(None, context=Context.EVAL)
def as_numeric(x: Any, _keep_na: bool = True) -> NumericOrIter:
    """Make elements numeric

    Args:
        x: The value to convert
        _keep_na: Whether to keep NAs as is. If True, will try to
            convert to double.

    Returns:
        Try `as_integer()` if failed then `as_float()`. If keep_na is True
    """
    if _keep_na:
        return as_double(x)
    try:
        return as_integer(x)
    except (ValueError, TypeError):
        return as_float(x)



"""Collects datasets from R-datasets, dplyr and tidyr packages"""
import functools
from pathlib import Path
from typing import List

import pandas
import toml

HERE = Path(__file__).parent


@functools.lru_cache()
def list_datasets():
    """Get the information of all datasets"""
    with HERE.joinpath("metadata.toml") as fmd:
        return toml.load(fmd)


@functools.lru_cache()
def load_data(name: str) -> pandas.DataFrame:
    """Load the specific dataset"""
    datasets = list_datasets()
    try:
        metadata = datasets[name]
    except KeyError:
        raise ImportError(
            f"No such dataset: {name}, "
            f"available: {list(datasets)}"
        ) from None

    data = pandas.read_csv(
        HERE / metadata["source"],
        index_col=0 if metadata["index"] else False,
    )
    data.__dfname__ = name
    return data


__all__ = []  # type: List[str]


def __getattr__(name):
    # mkapi accesses quite a lot of attributes starting with _
    try:
        return load_data(name.lower())
    except ImportError as imerr:
        raise AttributeError from imerr



"""Subset rows using column values

See source https://github.com/tidyverse/dplyr/blob/master/R/filter.R
"""
from typing import Iterable

import numpy
from pandas import DataFrame, RangeIndex
from pipda import register_verb
from pipda.expression import Expression
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.grouped import DataFrameGroupBy, DataFrameRowwise
from ..core.types import is_scalar, is_null, BoolOrIter
from ..core.utils import copy_attrs, reconstruct_tibble, Array
from .group_data import group_data, group_vars


@register_verb(DataFrame, context=Context.EVAL)
def filter(
    _data: DataFrame,
    *conditions: Iterable[bool],
    _preserve: bool = False,
    _drop_index: bool = None,
) -> DataFrame:
    """Subset a data frame, retaining all rows that satisfy your conditions

    Args:
        *conditions: Expressions that return logical values
        _preserve: Relevant when the .data input is grouped.
            If _preserve = FALSE (the default), the grouping structure
            is recalculated based on the resulting data, otherwise
            the grouping is kept as is.
        _drop_index: Whether drop the index or not.
            When it is None and the index of _data is a RangeIndex, then
            the index is dropped.

    Returns:
        The subset dataframe
    """
    if _data.shape[0] == 0 or not conditions:
        return _data

    condition = None
    for cond in conditions:
        cond = _sanitize_condition(cond, _data.shape[0])
        condition = (
            cond if condition is None else numpy.logical_and(condition, cond)
        )

    out = _data.loc[condition, :]
    if _drop_index is None:
        _drop_index = isinstance(_data.index, RangeIndex)

    if _drop_index:
        out = out.reset_index(drop=True)
    copy_attrs(out, _data)
    return out


@filter.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    _data: DataFrameGroupBy,
    *conditions: Expression,
    _preserve: bool = False,
    _drop_index: bool = None,  # TODO?
) -> DataFrameGroupBy:
    """Filter on DataFrameGroupBy object"""
    if _data.shape[0] > 0:
        out = _data._datar_apply(
            filter,
            *conditions,
            _drop_index=False,
            __calling_env=CallingEnvs.REGULAR,
        ).sort_index()
    else:
        out = _data.copy()

    out = reconstruct_tibble(_data, out)
    gdata = _filter_groups(out, _data)

    if not _preserve and _data.attrs["_group_drop"]:
        out.attrs["_group_data"] = gdata[
            gdata["_rows"].map(len) > 0
        ].reset_index(drop=True)

    return out


@filter.register(DataFrameRowwise, context=Context.EVAL)
def _(
    _data: DataFrameRowwise,
    *conditions: Expression,
    _preserve: bool = False,
    _drop_index: bool = None,
) -> DataFrameGroupBy:
    """Filter on DataFrameGroupBy object"""
    out = filter.dispatch(DataFrame)(
        _data, *conditions, _preserve=_preserve, _drop_index=_drop_index
    )
    return reconstruct_tibble(_data, out, keep_rowwise=True)


def _filter_groups(
    new: DataFrameGroupBy,
    old: DataFrameGroupBy,
    sort_rows: bool = True,
) -> DataFrame:
    """Filter non-existing rows in groupdata"""
    gdata = (
        group_data(new, __calling_env=CallingEnvs.REGULAR)
        .set_index(group_vars(new, __calling_env=CallingEnvs.REGULAR))["_rows"]
        .to_dict()
    )
    new_gdata = group_data(old, __calling_env=CallingEnvs.REGULAR).copy()
    for row in new_gdata.iterrows():
        ser = row[1]
        key = tuple(ser[:-1])
        if len(key) == 1:
            key = key[0]
        ser[-1] = gdata.get(key, [])
        new_gdata.loc[row[0], :] = ser

    if sort_rows:
        # GH69
        # The order changes when top row number filtered
        new_gdata = new_gdata.sort_values(
            ["_rows"],
            # keep empty rows at last
            key=lambda rowss: Array(
                [new.shape[0] if len(rows) == 0 else rows[0] for rows in rowss],
                dtype=int,
            ),
        ).reset_index(drop=True)

    new.attrs["_group_data"] = new_gdata
    return new_gdata


def _sanitize_condition(cond: BoolOrIter, length: int) -> numpy.ndarray:
    """Handle single condition"""
    if is_scalar(cond):
        out = Array([cond] * length)
    elif isinstance(cond, numpy.ndarray):
        out = cond
    else:
        out = Array(cond)
    out[is_null(out)] = False

    return out.astype(bool)



"""Functions related to complex numbers"""
from typing import Any

import numpy
from pandas.core.dtypes.common import is_complex_dtype
from pipda import register_func

from ..core.utils import register_numpy_func_x
from ..core.types import ComplexOrIter
from ..core.contexts import Context


from .testing import _register_type_testing
from .casting import _as_type

re = register_numpy_func_x(
    "re",
    "real",
    doc="""Real part of complex numbers

    Args:
        x: The complex numbers

    Returns:
        The real part of the complex numbers
    """,
)

im = register_numpy_func_x(
    "im",
    "imag",
    doc="""Imaginary part of complex numbers

    Args:
        x: The complex numbers

    Returns:
        The imaginary part of the complex numbers
    """,
)

mod = register_numpy_func_x(
    "mod",
    "absolute",
    doc="""Modulus of complex numbers

    Args:
        x: The complex numbers

    Returns:
        The Modulus of the complex numbers
    """,
)

arg = register_numpy_func_x(
    "arg",
    "angle",
    doc="""Angles of complex numbers

    Args:
        x: The complex numbers

    Returns:
        The Angles of the complex numbers
    """,
)

conj = register_numpy_func_x(
    "conj",
    "conj",
    doc="""conjugate of complex numbers

    Args:
        x: The complex numbers

    Returns:
        The conjugate of the complex numbers
    """,
)


@register_func(None, context=Context.EVAL)
def as_complex(x: Any, complex_type=numpy.complex_) -> ComplexOrIter:
    """Convert an object or elements of an iterable into complex

    Args:
        x: The object
        complex_type: The type of the complex. Could be one of:
            - numpy.complex128
            - numpy.complex256
            - numpy.complex64
            - numpy.complex_

    Returns:
        When x is an array or a series, return x.astype(bool).
        When x is iterable, convert elements of it into bools
        Otherwise, convert x to bool.
    """
    return _as_type(x, complex_type)


is_complex = _register_type_testing(
    "is_complex",
    scalar_types=(complex, numpy.complex_),
    dtype_checker=is_complex_dtype,
    doc="""Test if a value is complexes

    Args:
        x: The value to be checked

    Returns:
        True if the value is an complex or complexes; False otherwise.
    """,
)



"""Checking an iterable against itself or another one"""
from typing import Iterable

import numpy
from pipda import register_func

from ..core.utils import get_option
from ..core.contexts import Context


@register_func(None, context=Context.EVAL)
def which(x: Iterable[bool], base0_: bool = None) -> Iterable[int]:
    """Convert a bool iterable to indexes

    Args:
        x: An iterable of bools.
            Note that non-bool values will be converted into
        base0_: Whether the returned indexes are 0-based.
            Controlled by `get_option('which.base.0')` if not provided

    Returns:
        The indexes
    """
    return numpy.flatnonzero(x) + int(not get_option("which.base.0", base0_))


@register_func(None)
def which_min(x: Iterable, base0_: bool = None) -> int:
    """R's `which.min()`

    Get the index of the element with the maximum value

    Args:
        x: The iterable
        base0_: Whether the index to return is 0-based or not.
            Controlled by `get_option('which.base.0')` if not provided

    Returns:
        The index of the element with the maximum value
    """
    return numpy.argmin(x) + int(not get_option("which.base.0", base0_))


@register_func(None)
def which_max(x: Iterable, base0_: bool = True) -> int:
    """R's `which.max()`

    Get the index of the element with the minimum value

    Args:
        x: The iterable
        base0_: Whether the index to return is 0-based or not
            Not that this is not controlled by `get_option('index.base.0')`

    Returns:
        The index of the element with the minimum value
    """
    return numpy.argmax(x) + int(not get_option("which.base.0", base0_))
